#!/usr/bin/env python3

sway = False

i3 = None

outputs = {}
mon_desc2output_name = {}  # {'Samsung Electric Company SyncMaster 0x4B493234': 'HDMI-A-1', (...)}
outputs_num = 0
windows_list = []
h_taskbars_list = []
h_workspaces_list = []
h_submaps_list = []
niri_taskbars_list = []
controls_list = []
executors_list = []
tray_list = []
config_dir = ""
dwl_data_file = None
dwl_instances = []
name2icon_dict = {}
scratchpad_cons = {}

commands = {
    "light": False,
    "brightnessctl": False,
    "ddcutil": False,
    "pamixer": False,
    "pactl": False,
    "playerctl": False,
    "wlr-randr": False,
    "hyprctl": False,
    "upower": False,
    "swaync": False,
    "nwg-notifications": False
}

icons_path = ""  # "icons_light", "icons_dark" or "" (GTK icons)

defaults = {}



#!/usr/bin/env python3

from gi.repository import Gtk, Gdk, GLib, GtkLayerShell

import os
import subprocess
import random
import requests
import threading

from shutil import copyfile
from nwg_panel.tools import update_image, local_dir, cmd_through_compositor, create_background_task, eprint, save_json, \
    load_json


def on_enter_notify_event(widget, event):
    widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
    widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)


def on_leave_notify_event(widget, event):
    widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
    widget.unset_state_flags(Gtk.StateFlags.SELECTED)


class RandomWallpaper(Gtk.Button):
    def __init__(self, settings, voc, icons_path=""):
        self.image_info = {}
        defaults = {
            "source": "wallhaven.cc",
            "tags": ["nature"],
            "ratios": "16x9,16x10",
            "atleast": "1920x1080",
            "apikey": '',
            "save-path": os.getenv("HOME"),
            "local-path": "",
            "icon-size": 16,
            "interval": 0,
            "refresh-on-startup": True
        }
        for key in defaults:
            if key not in settings:
                settings[key] = defaults[key]
        self.settings = settings
        self.icons_path = icons_path
        self.wallpaper_path = os.path.join(local_dir(), "wallpaper.jpg")
        self.wallpaper_info_path = os.path.join(local_dir(), "wallpaper.json")

        self.voc = voc
        self.src_tag = 0

        Gtk.Button.__init__(self)
        self.set_always_show_image(True)
        self.settings = settings
        self.set_property("name", "random-wallpaper-button")

        image = Gtk.Image()
        update_image(image, "wallpaper", settings["icon-size"], icons_path)
        self.set_image(image)

        self.set_tooltip_text(voc["random-wallpaper"])
        self.connect('clicked', self.display_menu)

        self.connect('enter-notify-event', on_enter_notify_event)
        self.connect('leave-notify-event', on_leave_notify_event)

        self.show()

        if self.settings["refresh-on-startup"]:
            self.apply_wallpaper(None)

        if self.settings["interval"] > 0:
            self.src_tag = GLib.timeout_add_seconds(self.settings["interval"] * 60, self.apply_wallpaper, None)
            image = Gtk.Image()
            update_image(image, "wallpaper-refresh", settings["icon-size"], icons_path)
            self.set_image(image)

    def load_wallhaven_image(self):
        api_key = self.settings['apikey'] if self.settings["apikey"] else None
        api_key_status = "set" if self.settings['apikey'] else "unset"
        tags = " ".join(self.settings['tags']) if self.settings['tags'] else ""
        ratios = self.settings['ratios'] if self.settings['ratios'] else "16x9,16x10"
        atleast = self.settings["atleast"] if self.settings["atleast"] else "1920x1080"
        print(
            f"Fetching random image from wallhaven.cc, tags: '{tags}', ratios: '{ratios}', atleast: '{atleast}', API key: {api_key_status}")
        # Wallhaven API endpoint
        url = "https://wallhaven.cc/api/v1/search"

        # Parameters for the API request
        params = {
            "q": tags,
            "ratios": ratios,
            "atleast": atleast,
            "sorting": "random"
        }
        if api_key:
            params["apikey"] = api_key

        # Make the API request
        response = requests.get(url, params=params)

        # Get the image URL from the response
        if response.status_code == 200:
            image_data = response.json()
            try:
                image_url = image_data["data"][0]["path"]
            except (IndexError, KeyError):
                msg = self.voc["no-wallpaper-found"]
                tags = ",".join(self.settings["tags"])
                subprocess.Popen(f"notify-send '{msg}' {tags} -i preferences-desktop-wallpaper -t 6000",
                                 shell=True)

            self.image_info = image_data["data"][0]

            # Download the image
            image_response = requests.get(image_url)

            if image_response.status_code == 200:
                # Save the image locally
                with open(self.wallpaper_path, "wb") as file:
                    file.write(image_response.content)
                print(f"Wallhaven image saved as {self.wallpaper_path}")
                save_json(self.image_info, self.wallpaper_info_path)
                print(f"Wallhaven image info saved as {self.wallpaper_info_path}")
            else:
                eprint("Failed to download Wallhaven image")
        else:
            eprint("Failed to fetch Wallhaven image")

    def load_and_apply_local_image(self):
        if not os.path.isdir(self.settings["local-path"]):
            eprint("Local wallpaper path does not exist")
            return
        paths = os.listdir(self.settings["local-path"])
        if len(paths) == 0:
            eprint(f"No files found in {self.settings['local-path']}")
            return
        idx = random.randint(0, len(paths) - 1)
        image_path = os.path.join(self.settings["local-path"], paths[idx])
        ext = image_path.split(".")[-1]
        if ext.upper() in ["PNG", "JPG", "JPEG", "TIF", "TIFF", "SVG", 'WEBP', 'HEIC', 'AVIF']:
            print(f"Setting '{image_path}' as wallpaper")
            cmd = "pkill swaybg"
            subprocess.Popen('{}'.format(cmd), shell=True)
            print(f"Executing: {cmd}")
            subprocess.Popen('{}'.format(cmd), shell=True)

            cmd = f"swaybg -i '{image_path}' -m fill"

            cmd = cmd_through_compositor(cmd)
            print(f"Executing: {cmd}")
            subprocess.Popen(f"{cmd}", shell=True, preexec_fn=os.setpgrp)
        else:
            eprint(f"'{image_path}' is not a valid image file")

    def load_apply_wallhaven_image(self):
        self.load_wallhaven_image()

        cmd = "pkill swaybg"
        print(f"Executing: {cmd}")
        subprocess.Popen('{}'.format(cmd), shell=True)

        cmd = "swaybg -i {} -m fill".format(self.wallpaper_path)

        if os.path.isfile(self.wallpaper_path):
            cmd = cmd_through_compositor(cmd)
            print(f"Executing: {cmd}")
            subprocess.Popen('{}'.format(cmd), shell=True)
        else:
            eprint(f"'{self.wallpaper_path}' image not found")

    def apply_wallpaper(self, widget):
        if self.settings["source"] == "local":
            if os.path.isdir(self.settings["local-path"]) and len(os.listdir(self.settings["local-path"])) > 0:
                self.load_and_apply_local_image()
            else:
                eprint(f"Local wallpaper path {self.settings['local-path']} not found or empty")
        else:
            thread = threading.Thread(target=self.load_apply_wallhaven_image, daemon=True)
            thread.start()

        return True

    def display_menu(self, button):
        menu = Gtk.Menu()
        menu.set_reserve_toggle_size(False)

        item = Gtk.MenuItem.new_with_label(self.voc["refresh"])
        item.connect("activate", self.apply_and_reset_timer)
        menu.append(item)

        if self.settings["interval"] > 0:
            if self.src_tag > 0:
                item = Gtk.MenuItem.new_with_label(self.voc['pause-refresh'])
                item.connect("activate", self.pause_refresh)
            else:
                item = Gtk.MenuItem.new_with_label(self.voc['resume-refresh'])
                item.connect("activate", self.resume_refresh)
            menu.append(item)

        if self.settings["source"] == "wallhaven.cc":
            item = Gtk.MenuItem.new_with_label(self.voc["image-info"])
            item.connect("activate", self.display_image_info_window)
            menu.append(item)

            item = Gtk.MenuItem.new_with_label(self.voc["random-wallpaper-save"])
            item.connect("activate", self.save_wallpaper)
            menu.append(item)

        menu.show_all()
        menu.popup_at_widget(self, Gdk.Gravity.NORTH, Gdk.Gravity.SOUTH, None)

        on_leave_notify_event(button, None)

    def pause_refresh(self, item):
        if self.src_tag > 0:
            GLib.Source.remove(self.src_tag)
            self.src_tag = 0

            image = Gtk.Image()
            update_image(image, "wallpaper", self.settings["icon-size"], self.icons_path)
            self.set_image(image)

    def resume_refresh(self, item):
        self.src_tag = GLib.timeout_add_seconds(self.settings["interval"] * 60, self.apply_wallpaper, None)

        image = Gtk.Image()
        update_image(image, "wallpaper-refresh", self.settings["icon-size"], self.icons_path)
        self.set_image(image)

    def apply_and_reset_timer(self, btn):
        self.apply_wallpaper(None)
        if self.settings["interval"] > 0:
            GLib.Source.remove(self.src_tag)
            self.src_tag = GLib.timeout_add_seconds(self.settings["interval"] * 60, self.apply_wallpaper, None)

    def display_image_info_window(self, item):
        w = ImageInfoWindow(self.image_info, self.voc)

    def save_wallpaper(self, item):
        info = load_json(self.wallpaper_info_path)
        output_file_name = f"wallhaven-{info['id']}.jpg"
        save_path = self.settings["save-path"] if self.settings["save-path"] else os.getenv("HOME")
        try:
            msg = f"{self.voc['saved-to']} {save_path}"
            copyfile(self.wallpaper_path, os.path.join(save_path, output_file_name))
            subprocess.Popen(f"notify-send '{output_file_name}' '{msg}' -i preferences-desktop-wallpaper", shell=True)
        except Exception as e:
            eprint(f"Failed saving: {os.path.join(save_path, output_file_name)}: {e}")


class ImageInfoWindow(Gtk.Window):
    def __init__(self, image_info, voc):
        Gtk.Window.__init__(self, type_hint=Gdk.WindowTypeHint.NORMAL)
        GtkLayerShell.init_for_window(self)
        GtkLayerShell.set_namespace(self, "nwg-panel")
        GtkLayerShell.set_keyboard_mode(self, GtkLayerShell.KeyboardMode.ON_DEMAND)

        self.connect("key-release-event", self.handle_keyboard)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        self.add(vbox)

        lbl = Gtk.Label()
        lbl.set_markup(f"<b>{voc['random-wallhaven-wallpaper']}</b>")
        vbox.pack_start(lbl, False, False, 6)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        vbox.pack_start(hbox, False, False, 6)

        lines = ""
        for key in image_info:
            value = image_info[key]
            # nested dicts
            if isinstance(value, dict):
                l = ""
                for k in value:
                    l += f"\n\t{str(value[k])}"
                value = l
            if key == "url" or key == "short_url" or key == "path":
                value = f"<a href='{value}'>{value}</a>"
            line = f"<b>{key}</b>: {value}"
            lines += f"{line}\n"

        lbl = Gtk.Label()
        lbl.set_markup(lines)

        hbox.pack_start(lbl, False, False, 12)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        vbox.pack_start(hbox, False, False, 6)
        btn = Gtk.Button()
        btn.set_label(voc["close"])
        btn.connect("clicked", self.quit)
        hbox.pack_end(btn, False, False, 6)

        self.show_all()

    def handle_keyboard(self, widget, event):
        if event.type == Gdk.EventType.KEY_RELEASE and event.keyval == Gdk.KEY_Escape:
            self.destroy()

    def quit(self, btn):
        self.destroy()



#!/usr/bin/env python3

import time
import os
import subprocess

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
gi.require_version('GtkLayerShell', '0.1')
from gi.repository import Gtk, Gdk, GLib, GtkLayerShell

from nwg_panel.tools import (check_key, get_brightness, set_brightness, get_volume, get_balance, set_volume, get_battery,
                             update_image, eprint, list_sinks, toggle_mute, create_background_task, list_sink_inputs,
                             is_command, cmd_through_compositor)

from nwg_panel.common import commands

bat_critical_last_check = 0


class Controls(Gtk.EventBox):
    def __init__(self, settings, position, alignment, width, monitor=None, icons_path=""):
        self.settings = settings
        self.position = position
        self.alignment = alignment
        self.icons_path = icons_path
        Gtk.EventBox.__init__(self)

        defaults = {
            "show-brightness": False,
            "show-volume": False,
            "show-battery": True,
            "icon-size": 16,
            "interval": 1,
            "hover-opens": True,
            "leave-closes": False,
            "click-closes": True,
            "root-css-name": "controls-overview",
            "components": ["brightness", "battery", "volume", "readme", "processes"],
            "angle": 0,
            "battery-low-level": 20,
            "battery-low-interval": 3,
            "readme-label": "README",
            "processes-label": "Processes"
        }
        for key in defaults:
            check_key(settings, key, defaults[key])

        self.set_property("name", settings["root-css-name"])

        self.icon_size = settings["icon-size"]

        self.bri_icon_name = "view-refresh-symbolic"
        self.bri_image = Gtk.Image.new_from_icon_name(self.bri_icon_name, Gtk.IconSize.MENU)
        self.bri_label = Gtk.Label() if settings["show-brightness"] else None
        if self.bri_label:
            self.bri_label.set_property("name", "executor-label")
        self.bri_value = 0

        self.vol_icon_name = "view-refresh-symbolic"
        self.vol_image = Gtk.Image.new_from_icon_name(self.vol_icon_name, Gtk.IconSize.MENU)
        self.vol_label = Gtk.Label() if settings["show-volume"] else None
        if self.vol_label:
            self.vol_label.set_property("name", "executor-label")
        self.vol_value = 0
        self.vol_muted = False

        self.bat_icon_name = "view-refresh-symbolic"
        self.bat_image = Gtk.Image.new_from_icon_name(self.bat_icon_name, Gtk.IconSize.MENU)
        self.bat_label = Gtk.Label() if settings["show-battery"] else None
        if self.bat_label:
            self.bat_label.set_property("name", "executor-label")
        self.bat_value = 0
        self.bat_time = ""
        self.bat_charging = False

        self.pan_image = Gtk.Image()
        update_image(self.pan_image, "pan-down-symbolic", self.icon_size, self.icons_path)

        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.add(self.box)

        self.popup_window = PopupWindow(self, position, alignment, settings, width, monitor=monitor,
                                        icons_path=self.icons_path)

        self.connect('button-release-event', self.on_button_release, settings)
        self.connect('enter-notify-event', self.on_enter_notify_event, settings)
        self.connect('leave-notify-event', self.on_leave_notify_event)

        self.build_box()
        self.refresh()

        if "battery" in settings["components"]:
            self.refresh_bat()

    def build_box(self):
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        if self.settings["angle"] != 0.0:
            box.set_orientation(Gtk.Orientation.VERTICAL)
        self.box.pack_start(box, False, False, 6)

        if "brightness" in self.settings["components"]:
            box.pack_start(self.bri_image, False, False, 4)
            if self.bri_label:
                box.pack_start(self.bri_label, False, False, 0)

        if "volume" in self.settings["components"]:
            box.pack_start(self.vol_image, False, False, 4)
            if self.vol_label:
                box.pack_start(self.vol_label, False, False, 0)

        if "battery" in self.settings["components"]:
            box.pack_start(self.bat_image, False, False, 4)
            if self.bat_label:
                box.pack_start(self.bat_label, False, False, 0)

        box.pack_start(self.pan_image, False, False, 4)

    def refresh_output(self):
        if "brightness" in self.settings["components"]:
            try:
                self.bri_value = get_brightness(
                    device=self.settings["backlight-device"],
                    controller=self.settings["backlight-controller"])
                GLib.idle_add(self.update_brightness)
            except Exception as e:
                eprint(e)

        if "volume" in self.settings["components"] and (commands["pamixer"] or commands["pactl"]):
            try:
                GLib.idle_add(self.update_volume)
            except Exception as e:
                print(e)

    def refresh_bat_output(self):
        if "battery" in self.settings["components"]:
            try:
                self.bat_value, self.bat_time, self.bat_charging = get_battery()
                GLib.idle_add(self.update_battery, self.bat_value, self.bat_charging)
            except Exception as e:
                print(e)

    def refresh(self):
        thread = create_background_task(self.refresh_output, self.settings["interval"])
        thread.start()

    # No point in checking battery data more often that every 5 seconds
    def refresh_bat(self):
        thread = create_background_task(self.refresh_bat_output, 5)
        thread.start()

    def update_brightness(self, get=True):
        icon_name = bri_icon_name(self.bri_value)

        if icon_name != self.bri_icon_name:
            update_image(self.bri_image, icon_name, self.icon_size, self.icons_path)
            self.bri_icon_name = icon_name

        if self.bri_label:
            self.bri_label.set_text("{}%".format(self.bri_value))

        if get:
            self.popup_window.refresh()

    def update_volume(self):
        volume = get_volume()
        if (self.vol_value, self.vol_muted != volume):
            icon_name = vol_icon_name(*volume)

            if icon_name != self.vol_icon_name:
                update_image(self.vol_image, icon_name, self.settings["icon-size"], self.icons_path)
                self.vol_icon_name = icon_name

            if self.vol_label:
                self.vol_label.set_text("{}%".format(volume[0]))

            self.vol_value, self.vol_muted = volume

    def update_battery(self, value, charging):
        icon_name = bat_icon_name(value, charging)

        if icon_name != self.bat_icon_name:
            update_image(self.bat_image, icon_name, self.icon_size, self.icons_path)
            self.bat_icon_name = icon_name

        if self.bat_label:
            self.bat_label.set_text("{}%".format(value))

        if self.settings["battery-low-interval"] > 0:
            t = int(time.time())
            global bat_critical_last_check
            if not charging and t - bat_critical_last_check >= self.settings["battery-low-interval"] * 60 and value < \
                    self.settings["battery-low-level"]:
                subprocess.Popen('notify-send "Battery low! ({}%)" -i {}'.format(value, icon_name), shell=True)
                bat_critical_last_check = t

    def on_button_release(self, w, event, settings):
        if not self.popup_window.get_visible():
            self.popup_window.show_all()
            if settings["click-closes"]:
                self.popup_window.bcg_window.show()
            if self.popup_window.sink_box:
                self.popup_window.sink_box.hide()
            if self.popup_window.menu_box:
                self.popup_window.menu_box.hide()
        else:
            self.popup_window.hide()
            if self.popup_window.bcg_window:
                self.popup_window.bcg_window.hide()
        return False

    def on_enter_notify_event(self, widget, event, settings):
        if self.settings["hover-opens"]:
            if not self.popup_window.get_visible():
                self.popup_window.show_all()
                if settings["click-closes"]:
                    self.popup_window.bcg_window.show()
                if self.popup_window.sink_box:
                    self.popup_window.sink_box.hide()
                if self.popup_window.menu_box:
                    self.popup_window.menu_box.hide()
        else:
            widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
            widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)

        # cancel popup window close, as it's probably unwanted ATM
        self.popup_window.on_window_enter()

        return True

    def on_leave_notify_event(self, widget, event):
        widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
        widget.unset_state_flags(Gtk.StateFlags.SELECTED)
        return True


class PopupWindow(Gtk.Window):
    def __init__(self, parent, position, alignment, settings, width, monitor=None, icons_path=""):
        Gtk.Window.__init__(self, type_hint=Gdk.WindowTypeHint.NORMAL)
        self.balance = None
        self.bcg_window = None
        GtkLayerShell.init_for_window(self)
        if monitor:
            GtkLayerShell.set_monitor(self, monitor)

        check_key(settings, "backlight-controller", "brightnessctl")
        check_key(settings, "backlight-device", "")

        check_key(settings, "css-name", "controls-window")
        self.parent = parent

        self.set_up_bcg_window()

        self.set_property("name", settings["css-name"])
        self.icon_size = settings["icon-size"]
        self.icons_path = icons_path

        self.settings = settings
        self.position = position

        self.menu_box = None
        self.sink_box = None

        self.bri_scale = None
        self.bri_scale_handler = None
        self.vol_scale = None
        self.vol_scale_handler = None

        self.per_app_sliders = []

        self.src_tag = 0

        self.connect("show", self.on_window_show)

        check_key(settings, "output-switcher", False)
        self.sinks = []
        if (commands["pamixer"] or commands["pactl"]) and settings["output-switcher"]:
            self.sinks = list_sinks()

        eb = Gtk.EventBox()
        eb.set_above_child(False)
        if settings["leave-closes"]:
            self.connect("leave_notify_event", self.on_window_exit)
            self.connect("enter_notify_event", self.on_window_enter)

        outer_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        eb.add(outer_vbox)
        self.add(eb)

        outer_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        Gtk.Widget.set_size_request(outer_hbox, width, 10)
        outer_vbox.pack_start(outer_hbox, True, True, 20)

        v_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        outer_hbox.pack_start(v_box, True, True, 20)

        GtkLayerShell.set_layer(self, GtkLayerShell.Layer.OVERLAY)

        check_key(settings, "window-margin-vertical", 0)
        if position == "top":
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.TOP, settings["window-margin-vertical"])
        else:
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.BOTTOM, settings["window-margin-vertical"])

        check_key(settings, "window-margin-horizontal", 0)
        if alignment == "right":
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.RIGHT, settings["window-margin-horizontal"])
        else:
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.LEFT, settings["window-margin-horizontal"])

        # Since v0.7 alignment 'left' / 'right' has been renamed to 'start' / 'end' in the GUI,
        # but will stay as is in the code. We also have 4 positions now: 'top', 'bottom', 'left', 'right'
        # instead of 'top" & 'bottom'.
        if alignment == "left":  # Means: 'start'
            if position == "top":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.TOP, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.LEFT, True)
            elif position == "bottom":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.BOTTOM, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.LEFT, True)
            elif position == "left":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.LEFT, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.TOP, True)
            elif position == "right":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.RIGHT, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.TOP, True)
        else:  # align 'end'
            if position == "top":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.TOP, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.RIGHT, True)
            elif position == "bottom":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.BOTTOM, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.RIGHT, True)
            elif position == "left":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.LEFT, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.BOTTOM, True)
            elif position == "right":
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.RIGHT, True)
                GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.BOTTOM, True)

        check_key(settings, "commands", {"battery": ""})

        add_sep = False
        if "brightness" in settings["components"]:
            self.value_changed = False
            self.scrolled = False

            inner_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
            v_box.pack_start(inner_hbox, False, False, 0)

            self.bri_icon_name = "view-refresh-symbolic"
            self.bri_image = Gtk.Image.new_from_icon_name(self.bri_icon_name, Gtk.IconSize.MENU)

            inner_hbox.pack_start(self.bri_image, False, False, 6)

            self.bri_scale = Gtk.Scale.new_with_range(orientation=Gtk.Orientation.HORIZONTAL, min=0, max=100, step=1)
            self.bri_scale.set_increments(0.5, 0.5)

            if self.settings["backlight-controller"] == "ddcutil":
                self.bri_scale_handler = self.bri_scale.connect("value-changed", self.on_value_changed)
                self.bri_scale.connect("button-release-event", self.on_button_release)

                self.bri_scale.add_events(Gdk.EventMask.SCROLL_MASK)
                self.bri_scale.connect('scroll-event', self.on_scroll)
            else:
                self.bri_scale_handler = self.bri_scale.connect("value-changed", self.set_bri)

            inner_hbox.pack_start(self.bri_scale, True, True, 5)
            add_sep = True

        if "volume" in settings["components"] and (commands["pamixer"] or commands["pactl"]):
            inner_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
            v_box.pack_start(inner_hbox, False, False, 6)

            self.vol_icon_name = "view-refresh-symbolic"
            self.vol_image = Gtk.Image.new_from_icon_name(self.vol_icon_name, Gtk.IconSize.MENU)

            if self.parent.vol_icon_name != self.vol_icon_name:
                update_image(self.vol_image, self.parent.vol_icon_name, self.icon_size, self.icons_path)
                self.vol_icon_name = self.parent.vol_icon_name

            eb = Gtk.EventBox()
            eb.connect("enter_notify_event", self.on_enter_notify_event)
            eb.connect("leave_notify_event", self.on_leave_notify_event)
            eb.connect("button-release-event", self.toggle_mute)
            eb.add(self.vol_image)
            inner_hbox.pack_start(eb, False, False, 6)

            self.vol_scale = Gtk.Scale.new_with_range(orientation=Gtk.Orientation.HORIZONTAL, min=0, max=100, step=1)
            self.vol_scale.set_increments(0.5, 0.5)
            self.vol_scale.set_value(self.parent.vol_value)
            self.vol_scale_handler = self.vol_scale.connect("value-changed", self.set_vol)

            inner_hbox.pack_start(self.vol_scale, True, True, 5)
            if (commands["pamixer"] or commands["pactl"]) and settings["output-switcher"]:
                pactl_eb = Gtk.EventBox()
                image = Gtk.Image()
                pactl_eb.add(image)
                pactl_eb.connect("enter_notify_event", self.on_enter_notify_event)
                pactl_eb.connect("leave_notify_event", self.on_leave_notify_event)
                update_image(image, "pan-down-symbolic", self.icon_size, self.icons_path)
                inner_hbox.pack_end(pactl_eb, False, False, 5)

                self.sink_box = SinkBox()
                pactl_eb.connect('button-release-event', self.sink_box.switch_visibility)
                v_box.pack_start(self.sink_box, False, False, 0)

            add_sep = True

        if "per-app-volume" in settings["components"] and commands["pactl"]:
            self.per_app_vol_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
            v_box.pack_start(self.per_app_vol_box, False, False, 10)
        else:
            self.per_app_vol_box = None

        if add_sep:
            sep = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
            v_box.pack_start(sep, True, True, 10)

        if "battery" in settings["components"]:
            event_box = Gtk.EventBox()
            if "battery" in settings["commands"] and settings["commands"]["battery"]:
                event_box.connect("enter_notify_event", self.on_enter_notify_event)
                event_box.connect("leave_notify_event", self.on_leave_notify_event)

                event_box.connect('button-release-event', self.launch, settings["commands"]["battery"])

            inner_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
            inner_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
            inner_vbox.pack_start(inner_hbox, True, True, 6)
            v_box.pack_start(event_box, True, True, 0)

            self.bat_icon_name = "view-refresh-symbolic"
            self.bat_image = Gtk.Image.new_from_icon_name(self.bat_icon_name, Gtk.IconSize.MENU)

            inner_hbox.pack_start(self.bat_image, False, False, 6)

            self.bat_label = Gtk.Label()
            inner_hbox.pack_start(self.bat_label, False, True, 6)

            if "battery" in settings["commands"] and settings["commands"]["battery"]:
                img = Gtk.Image()
                update_image(img, "pan-end-symbolic", self.icon_size, self.icons_path)
                inner_hbox.pack_end(img, False, True, 4)

            event_box.add(inner_vbox)

        if "readme" in settings["components"] and is_command("nwg-readme-browser"):
            event_box = Gtk.EventBox()
            event_box.connect("enter_notify_event", self.on_enter_notify_event)
            event_box.connect("leave_notify_event", self.on_leave_notify_event)
            event_box.connect('button-release-event', self.launch, "nwg-readme-browser -i")

            inner_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
            inner_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
            inner_vbox.pack_start(inner_hbox, True, True, 6)
            v_box.pack_start(event_box, True, True, 0)

            self.proc_image = Gtk.Image()
            update_image(self.proc_image, "nwg-readme-browser", self.icon_size, self.icons_path)

            inner_hbox.pack_start(self.proc_image, False, False, 6)

            self.proc_label = Gtk.Label.new(settings["readme-label"])
            inner_hbox.pack_start(self.proc_label, False, True, 6)

            img = Gtk.Image()
            update_image(img, "pan-end-symbolic", self.icon_size, self.icons_path)
            inner_hbox.pack_end(img, False, True, 4)

            event_box.add(inner_vbox)

        if "processes" in settings["components"]:
            event_box = Gtk.EventBox()
            event_box.connect("enter_notify_event", self.on_enter_notify_event)
            event_box.connect("leave_notify_event", self.on_leave_notify_event)
            event_box.connect('button-release-event', self.launch, "nwg-processes")

            inner_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
            inner_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
            inner_vbox.pack_start(inner_hbox, True, True, 6)
            v_box.pack_start(event_box, True, True, 0)

            self.proc_image = Gtk.Image()
            update_image(self.proc_image, "nwg-processes", self.icon_size, self.icons_path)

            inner_hbox.pack_start(self.proc_image, False, False, 6)

            self.proc_label = Gtk.Label.new(settings["processes-label"])
            inner_hbox.pack_start(self.proc_label, False, True, 6)

            img = Gtk.Image()
            update_image(img, "pan-end-symbolic", self.icon_size, self.icons_path)
            inner_hbox.pack_end(img, False, True, 4)

            event_box.add(inner_vbox)

        check_key(settings, "custom-items", [])
        if settings["custom-items"]:
            for item in settings["custom-items"]:
                check_key(item, "name", "undefined")
                check_key(item, "icon", "")
                check_key(item, "cmd", "")
                c_item = self.custom_item(item["name"], item["icon"], item["cmd"])
                v_box.pack_start(c_item, True, True, 2)

        check_key(settings, "menu", {})
        if settings["menu"] and "items" in settings["menu"] and settings["menu"]["items"]:
            template = settings["menu"]

            if "custom-items" in settings and len(settings["custom-items"]) > 0:
                sep = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
                v_box.pack_start(sep, True, True, 10)

            e_box = Gtk.EventBox()
            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
            e_box.add(box)
            inner_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
            box.pack_start(inner_hbox, True, True, 6)
            v_box.pack_start(e_box, True, True, 0)

            img = Gtk.Image()
            update_image(img, template["icon"], self.icon_size, self.icons_path)
            inner_hbox.pack_start(img, False, False, 6)

            check_key(template, "name", "Menu name")
            label = Gtk.Label(template["name"])
            inner_hbox.pack_start(label, False, False, 6)

            check_key(template, "items", [])
            if template["items"]:
                img = Gtk.Image()
                update_image(img, "pan-down-symbolic", self.icon_size, self.icons_path)
                inner_hbox.pack_end(img, False, True, 5)

                e_box.connect("enter-notify-event", self.on_enter_notify_event)
                e_box.connect("leave-notify-event", self.on_leave_notify_event)

                self.menu_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
                v_box.pack_start(self.menu_box, False, False, 0)
                for item in template["items"]:
                    eb = Gtk.EventBox()
                    vb = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
                    hb = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
                    vb.pack_start(hb, False, False, 3)
                    i = Gtk.Label(item["name"])
                    hb.pack_start(i, False, False, self.icon_size + 18)
                    eb.add(vb)
                    eb.connect("enter_notify_event", self.on_enter_notify_event)
                    eb.connect("leave_notify_event", self.on_leave_notify_event)
                    eb.connect("button-release-event", self.launch, item["cmd"])
                    self.menu_box.pack_start(eb, False, False, 0)

                e_box.connect('button-release-event', self.switch_menu_box)

        self.refresh(True)

    def get_balance(self):
        self.balance = get_balance()
        print("Audio balance", self.balance)

    def schedule_refresh(self):
        Gdk.threads_add_timeout(GLib.PRIORITY_LOW, 500, self.refresh, (True,))

    def set_up_bcg_window(self):
        self.bcg_window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)

        GtkLayerShell.init_for_window(self.bcg_window)
        GtkLayerShell.set_layer(self.bcg_window, GtkLayerShell.Layer.TOP)
        GtkLayerShell.set_anchor(self.bcg_window, GtkLayerShell.Edge.TOP, True)
        GtkLayerShell.set_anchor(self.bcg_window, GtkLayerShell.Edge.BOTTOM, True)
        GtkLayerShell.set_anchor(self.bcg_window, GtkLayerShell.Edge.LEFT, True)
        GtkLayerShell.set_anchor(self.bcg_window, GtkLayerShell.Edge.RIGHT, True)
        # GtkLayerShell.set_exclusive_zone(self.bcg_window, -1)  # cover panels
        self.bcg_window.connect("button-release-event", self.hide_and_clear_tag)
        self.bcg_window.set_property("name", "bcg-window")

    def on_window_exit(self, w, e):
        if self.get_visible():
            self.src_tag = GLib.timeout_add_seconds(1, self.hide_and_clear_tag)
        return True

    def hide_and_clear_tag(self, *args):
        self.hide()
        self.bcg_window.hide()
        self.src_tag = 0

    def on_window_enter(self, *args):
        if self.src_tag > 0:
            GLib.Source.remove(self.src_tag)
            self.src_tag = 0
        return True

    def on_window_show(self, *args):
        self.src_tag = 0
        if "per-app-volume" in self.settings["components"] and commands["pactl"]:
            self.create_per_app_sliders()
        self.get_balance()
        self.refresh()

    def create_per_app_sliders(self):
        self.per_app_sliders = []
        sink_inputs = list_sink_inputs()
        if self.per_app_vol_box:
            for c in self.per_app_vol_box.get_children():
                c.destroy()
            for inp in sink_inputs:
                props = sink_inputs[inp]["Properties"] if "Properties" in sink_inputs[inp]["Properties"] else {}
                if props:
                    icon_name = props[
                        "application.icon_name"] if "application.icon_name" in props else "emblem-music-symbolic"
                    vol = 0
                    if "Volume" in sink_inputs[inp]:
                        for p in sink_inputs[inp]["Volume"].split():
                            if p.endswith("%"):
                                try:
                                    vol = int(p[:-1])
                                except:
                                    pass
                        if "application.name" in props and "media.name" in props:
                            scale = PerAppSlider(inp, vol, icon_name, props["application.name"], props["media.name"])
                            self.per_app_sliders.append(scale)
                            self.per_app_vol_box.pack_start(scale, False, False, 0)
                            self.show_all()

    def switch_menu_box(self, widget, event):
        if self.menu_box.get_visible():
            self.menu_box.hide()
        else:
            self.menu_box.show_all()

    def refresh_sinks(self, *args):
        if commands["pamixer"] or commands["pactl"]:
            self.sinks = list_sinks()

    def toggle_mute(self, e, slider):
        toggle_mute()
        self.refresh()

    def custom_item(self, name, icon, cmd):
        eb = Gtk.EventBox()

        v_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        h_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        v_box.pack_start(h_box, False, False, 6)
        eb.add(v_box)

        image = Gtk.Image()
        update_image(image, icon, self.icon_size, self.icons_path)
        h_box.pack_start(image, False, True, 6)

        label = Gtk.Label(name)
        h_box.pack_start(label, False, True, 4)

        if cmd:
            eb.connect("enter_notify_event", self.on_enter_notify_event)
            eb.connect("leave_notify_event", self.on_leave_notify_event)
            eb.connect('button-release-event', self.launch, cmd)

            img = Gtk.Image()
            update_image(img, "pan-end-symbolic", self.icon_size, self.icons_path)
            h_box.pack_end(img, False, True, 4)

        return eb

    def refresh(self, schedule=False):
        if self.get_visible():
            self.refresh_sinks()
            # self.parent.refresh_output()

            if "battery" in self.settings["components"]:
                if self.parent.bat_icon_name != self.bat_icon_name:
                    update_image(self.bat_image, self.parent.bat_icon_name, self.icon_size, self.icons_path)
                    self.bat_icon_name = self.parent.bat_icon_name

                self.bat_label.set_text("{}% {}".format(self.parent.bat_value, self.parent.bat_time))

            if "volume" in self.settings["components"] and (commands["pamixer"] or commands["pactl"]):
                self.vol_scale.set_value(self.parent.vol_value)
                if self.parent.vol_icon_name != self.vol_icon_name:
                    update_image(self.vol_image, self.parent.vol_icon_name, self.icon_size, self.icons_path)
                    self.vol_icon_name = self.parent.vol_icon_name
                self.vol_scale.set_draw_value(
                    False if self.parent.vol_value > 100 else True)  # Don't display val out of scale

            if "per-app-volume" in self.settings["components"] and commands["pactl"]:
                # list input numbers we already have a slider for
                already_have_slider = []
                for s in self.per_app_sliders:
                    already_have_slider.append(str(s.input_num))

                sink_inputs = list_sink_inputs()
                inp_nums = []
                for inp in sink_inputs:
                    inp_nums.append(inp)

                for inp in sink_inputs:
                    if inp not in already_have_slider:
                        # We have no slider for input {inp}. Let's add it.
                        props = sink_inputs[inp]["Properties"] if "Properties" in sink_inputs[inp] else {}
                        if props:
                            icon_name = props[
                                "application.icon_name"] if "application.icon_name" in props else "emblem-music-symbolic"
                            vol = 0
                            if "Volume" in sink_inputs[inp]:
                                for p in sink_inputs[inp]["Volume"].split():
                                    if p.endswith("%"):
                                        try:
                                            vol = int(p[:-1])
                                        except:
                                            pass
                                if "application.name" in props and "media.name" in props:
                                    scale = PerAppSlider(inp, vol, icon_name, props["application.name"],
                                                         props["media.name"])
                                    self.per_app_sliders.append(scale)
                                    self.per_app_vol_box.pack_start(scale, False, False, 0)
                                    scale.show_all()

                                vol = 0
                                for p in sink_inputs[inp]["Volume"].split():
                                    if p.endswith("%"):
                                        try:
                                            vol = int(p[:-1])
                                        except:
                                            pass

                                for sc in self.per_app_sliders:
                                    if sc.input_num == inp:
                                        sc.scale.set_value(vol)

                # In case the app is closed while popup still open
                for sc in self.per_app_sliders:
                    if sc.input_num not in inp_nums:
                        sc.destroy()

            if "brightness" in self.settings["components"]:
                if not self.value_changed:
                    self.bri_scale.set_value(self.parent.bri_value)
                if self.parent.bri_icon_name != self.bri_icon_name:
                    update_image(self.bri_image, self.parent.bri_icon_name, self.icon_size, self.icons_path)
                    self.bri_icon_name = self.parent.bri_icon_name

        else:
            if "volume" in self.settings["components"] and (commands["pamixer"] or commands["pactl"]):
                with self.vol_scale.handler_block(self.vol_scale_handler):
                    self.vol_scale.set_value(self.parent.vol_value)

            if "brightness" in self.settings["components"]:
                with self.bri_scale.handler_block(self.bri_scale_handler):
                    self.bri_scale.set_value(self.parent.bri_value)

        if schedule:
            self.schedule_refresh()

    def on_enter_notify_event(self, widget, event):
        widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
        widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)
        return True

    def on_leave_notify_event(self, widget, event):
        widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
        widget.unset_state_flags(Gtk.StateFlags.SELECTED)
        return True

    def set_bri(self, slider):
        self.parent.bri_value = int(slider.get_value())
        self.parent.update_brightness(get=False)
        set_brightness(self.parent.bri_value, device=self.settings["backlight-device"],
                       controller=self.settings["backlight-controller"])

    def on_button_release(self, scale, event):
        if self.value_changed:
            self.set_bri(scale)
            self.value_changed = False

    def on_value_changed(self, *args):
        if self.scrolled:
            self.set_bri(self.bri_scale)
            self.scrolled = False
        else:
            self.value_changed = True

    def on_scroll(self, w, event):
        self.scrolled = True

    def set_vol(self, slider):
        self.parent.vol_value = int(slider.get_value())
        set_volume(self.parent.vol_value, balance=self.balance)
        # self.parent.update_volume()

    def close_win(self, w, e):
        self.hide()
        self.bcg_window.hide()

    def handle_keyboard(self, w, e):
        if e.type == Gdk.EventType.KEY_RELEASE and e.keyval == Gdk.KEY_Escape:
            self.close_win(w, e)
        return e

    def launch(self, w, e, cmd):
        cmd = cmd_through_compositor(cmd)
        print(f"Executing: {cmd}")
        subprocess.Popen('{}'.format(cmd), shell=True)
        self.hide()
        self.bcg_window.hide()


class PerAppSlider(Gtk.Box):
    def __init__(self, input_num, volume, icon_name, app_name, media_name):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
        self.set_property("margin-left", 6)
        self.set_property("margin-right", 6)
        self.input_num = input_num

        img = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU)
        self.pack_start(img, False, False, 0)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.pack_start(vbox, True, True, 0)
        name = f"{app_name}: {media_name}"
        if len(name) > 40:
            name = "{}…".format(name[:40])
        lbl = Gtk.Label()
        lbl.set_markup('<span size="small">{}</span>'.format(name))
        vbox.pack_start(lbl, False, False, 0)
        self.scale = Gtk.Scale.new_with_range(orientation=Gtk.Orientation.HORIZONTAL, min=0, max=153, step=1)
        self.scale.set_increments(0.5, 0.5)
        self.scale.connect("value-changed", self.set_volume)
        self.scale.set_value(volume)
        self.scale.set_draw_value(False)
        vbox.pack_start(self.scale, True, True, 0)

    def set_volume(self, scale):
        percent = scale.get_value()
        target = int(65536 * percent / 100)
        subprocess.Popen('exec pactl set-sink-input-volume {} {}'.format(self.input_num, target), shell=True)


class SinkBox(Gtk.Box):
    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.sinks = None
        self.refresh()

    def refresh(self):
        for item in self.get_children():
            item.destroy()
        if commands["pamixer"] or commands["pactl"]:
            self.sinks = list_sinks()
            for sink in self.sinks:
                eb = Gtk.EventBox()
                eb.connect("enter_notify_event", self.on_enter_notify_event)
                eb.connect("leave_notify_event", self.on_leave_notify_event)
                eb.connect('button-release-event', self.switch_sink, sink["name"])
                vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
                hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
                vbox.pack_start(hbox, True, True, 4)
                desc = sink["desc"]
                if len(desc) > 26:
                    desc = "{}\u2026".format(desc[:26])
                if sink["running"]:
                    desc = f"✓ {desc}"
                label = Gtk.Label(desc)
                hbox.pack_start(label, True, True, 0)
                eb.add(vbox)
                self.pack_start(eb, False, False, 0)

    def switch_visibility(self, *args):
        if self.get_visible():
            self.hide()
        else:
            self.refresh()
            self.show_all()

    def on_enter_notify_event(self, widget, event):
        widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
        widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)

    def on_leave_notify_event(self, widget, event):
        widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
        widget.unset_state_flags(Gtk.StateFlags.SELECTED)

    def switch_sink(self, w, e, sink):
        if commands["pactl"]:
            eprint("Sink: '{}'".format(sink))
            subprocess.Popen('exec pactl set-default-sink "{}"'.format(sink), shell=True)
        else:
            eprint("Couldn't switch sinks, 'pactl' (libpulse) not found")
        self.hide()


def bri_icon_name(value):
    icon_name = "display-brightness-low-symbolic"
    if value > 70:
        icon_name = "display-brightness-high-symbolic"
    elif value > 30:
        icon_name = "display-brightness-medium-symbolic"

    return icon_name


def vol_icon_name(value, muted):
    icon_name = "audio-volume-muted-symbolic"
    if not muted:
        if value is not None:
            if value > 70:
                icon_name = "audio-volume-high-symbolic"
            elif value > 30:
                icon_name = "audio-volume-medium-symbolic"
            else:
                icon_name = "audio-volume-low-symbolic"
        else:
            icon_name = "audio-volume-muted-symbolic"

    return icon_name


def bat_icon_name(value, is_charging):
    icon_name = "battery-empty-symbolic"
    if is_charging:
        if value > 90:
            icon_name = "battery-full-charging-symbolic"
        elif value > 40:
            icon_name = "battery-good-charging-symbolic"
        elif value > 19:
            icon_name = "battery-low-charging-symbolic"
        else:
            icon_name = "battery-empty-charging-symbolic"
    else:
        if value > 90:
            icon_name = "battery-full-symbolic"
        elif value > 40:
            icon_name = "battery-good-symbolic"
        elif value > 19:
            icon_name = "battery-low-symbolic"

    return icon_name



#!/usr/bin/env python3

from gi.repository import Gtk, Gdk, GLib
from i3ipc import Event

import nwg_panel.common
from nwg_panel.tools import check_key, get_icon_name, update_image, update_image_fallback_desktop, load_autotiling


class SwayWorkspaces(Gtk.Box):
    def __init__(self, settings, i3, icons_path):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.settings = settings
        self.i3 = i3
        self.num_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.num_box.set_property("name", "sway-workspaces")
        self.ws_num2box = {}
        self.ws_num2lbl = {}
        self.name_label = Gtk.Label()
        self.name_label.set_property("name", "sway-workspaces-name")
        self.win_id = ""
        self.win_pid = None
        self.icon = Gtk.Image()
        self.icon.set_property("name", "sway-workspaces-icon")
        self.layout_icon = Gtk.Image()
        self.icons_path = icons_path
        self.autotiling = load_autotiling()
        self.build_box()
        self.refresh()
        self.subscribe()

    def subscribe(self):
        self.i3.on(Event.WINDOW, self.on_i3ipc_event)
        self.i3.on(Event.WORKSPACE, self.on_i3ipc_event)

    def build_box(self):
        check_key(self.settings, "numbers", [])
        check_key(self.settings, "custom-labels", [])
        check_key(self.settings, "focused-labels", [])
        check_key(self.settings, "show-icon", True)
        check_key(self.settings, "image-size", 16)
        check_key(self.settings, "show-name", True)
        check_key(self.settings, "name-length", 40)
        check_key(self.settings, "mark-autotiling", True)
        check_key(self.settings, "mark-content", True)
        check_key(self.settings, "hide-empty", False)
        check_key(self.settings, "show-layout", True)
        check_key(self.settings, "angle", 0.0)
        if self.settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)
            self.num_box.set_orientation(Gtk.Orientation.VERTICAL)

        # prevent from #142
        ws_num = -1
        if self.i3.get_tree().find_focused():
            ws_num, win_name, win_id, non_empty, win_layout, numbers = self.find_details()

        if len(self.settings["custom-labels"]) == 1:
            self.settings["custom-labels"] *= len(self.settings["numbers"])
        elif len(self.settings["custom-labels"]) != len(self.settings["numbers"]):
            self.settings["custom-labels"] = []

        if len(self.settings["focused-labels"]) == 1:
            self.settings["focused-labels"] *= len(self.settings["numbers"])
        elif len(self.settings["focused-labels"]) != len(self.settings["numbers"]):
            self.settings["focused-labels"] = []

        self.pack_start(self.num_box, False, False, 0)

        for idx, num in enumerate(self.settings["numbers"]):
            if num == str(ws_num) and self.settings["focused-labels"]:
                label = self.settings["focused-labels"][idx]
            elif self.settings["custom-labels"]:
                label = self.settings["custom-labels"][idx]
            else:
                label = str(num)

            eb, lbl = self.build_number(num, label)
            self.num_box.pack_start(eb, False, False, 0)

            if num == str(ws_num):
                eb.set_property("name", "task-box-focused")
            else:
                eb.set_property("name", "task-box")

        if self.settings["show-icon"]:
            self.pack_start(self.icon, False, False, 6)

        if self.settings["show-name"]:
            self.pack_start(self.name_label, False, False, 0)

        if self.settings["show-layout"]:
            self.pack_start(self.layout_icon, False, False, 6)

    def build_number(self, num, label):
        eb = Gtk.EventBox()
        eb.connect("enter_notify_event", self.on_enter_notify_event)
        eb.connect("leave_notify_event", self.on_leave_notify_event)
        eb.connect("button-release-event", self.on_click, num)
        eb.add_events(Gdk.EventMask.SCROLL_MASK)
        eb.connect('scroll-event', self.on_scroll)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        box.set_property("name", "sway-workspaces-item")
        if self.settings["angle"] != 0.0:
            box.set_orientation(Gtk.Orientation.VERTICAL)
        eb.add(box)

        if self.settings["mark-autotiling"]:
            try:
                at = int(num) in self.autotiling
            except:
                at = False
            autotiling = "a" if at in self.autotiling else ""
            lbl = Gtk.Label("{}{}".format(autotiling, label))
        else:
            lbl = Gtk.Label("{}".format(label))
        lbl.set_use_markup(True)
        if self.settings["angle"] != 0.0:
            lbl.set_angle(self.settings["angle"])
            self.name_label.set_angle(self.settings["angle"])

        self.ws_num2box[num] = eb
        self.ws_num2lbl[num] = lbl

        box.pack_start(lbl, False, False, 6)

        return eb, lbl

    def on_i3ipc_event(self, i3conn, event):
        GLib.idle_add(self.refresh, priority=GLib.PRIORITY_HIGH)

    def refresh(self):
        if self.i3.get_tree().find_focused():
            ws_num, win_name, win_id, non_empty, win_layout, numbers = self.find_details()

            if len(self.settings["numbers"]) > 0:
                numbers = self.settings["numbers"]

            if ws_num > 0:
                for num in self.ws_num2lbl:
                    self.ws_num2lbl[num].hide()

                for _idx, num in enumerate(numbers):
                    idx = None
                    if num in self.settings["numbers"]:
                        idx = self.settings["numbers"].index(num)
                    try:
                        int_num = int(num)
                    except:
                        int_num = 0

                    if idx is None:
                        text = str(num)
                    elif num == str(ws_num) and self.settings["focused-labels"]:
                        text = self.settings["focused-labels"][idx]
                    elif self.settings["custom-labels"]:
                        text = self.settings["custom-labels"][idx]
                    else:
                        text = str(num)

                    if num not in self.ws_num2lbl:
                        eb, lbl = self.build_number(num, text)
                        self.num_box.pack_start(eb, False, False, 0)
                        eb.show_all()

                    lbl = self.ws_num2lbl[num]

                    if not self.settings["hide-empty"] or int_num in non_empty or num == str(ws_num):
                        lbl.show()
                    else:
                        lbl.hide()

                    # mark non-empty WS with CSS ID
                    if int_num in non_empty:
                        lbl.set_property("name", "workspace-occupied")

                    # mark non-empty WS with a dot
                    if self.settings["mark-content"]:
                        if int_num in non_empty:
                            if not text.endswith("."):
                                text += "."
                        else:
                            if text.endswith("."):
                                text = text[0:-1]

                    lbl.set_markup(text)

                    if num == str(ws_num):
                        self.ws_num2box[num].set_property("name", "task-box-focused")
                    else:
                        self.ws_num2box[num].set_property("name", "task-box")

                if self.settings["show-icon"] and win_id != self.win_id:
                    self.update_icon(win_id, win_name)
                    self.win_id = win_id

            if self.settings["show-name"]:
                self.name_label.set_text(win_name)

            if self.settings["show-layout"]:
                if win_name:
                    if win_layout == "splith":
                        update_image(self.layout_icon, "go-next-symbolic", self.settings["image-size"], self.icons_path)
                    elif win_layout == "splitv":
                        update_image(self.layout_icon, "go-down-symbolic", self.settings["image-size"], self.icons_path)
                    elif win_layout == "tabbed":
                        update_image(self.layout_icon, "view-dual-symbolic", self.settings["image-size"],
                                     self.icons_path)
                    elif win_layout == "stacked":
                        update_image(self.layout_icon, "view-paged-symbolic", self.settings["image-size"],
                                     self.icons_path)
                    else:
                        update_image(self.layout_icon, "window-pop-out-symbolic", self.settings["image-size"],
                                     self.icons_path)

                    if not self.layout_icon.get_visible():
                        self.layout_icon.show()
                else:
                    if self.layout_icon.get_visible():
                        self.layout_icon.hide()

    def update_icon(self, win_id, win_name):
        loaded_icon = False
        if win_id and win_name:
            try:
                update_image_fallback_desktop(self.icon,
                                              win_id,
                                              self.settings["image-size"],
                                              self.icons_path,
                                              fallback=False)
                loaded_icon = True
                if not self.icon.get_visible():
                    self.icon.show()
            except:
                pass

        if not loaded_icon and self.icon.get_visible():
            self.icon.hide()

    def find_details(self):
        tree = self.i3.get_tree()
        workspaces = self.i3.get_workspaces()
        ws_num = -1
        win_name = ""
        win_id = ""  # app_id if available, else window_class
        layout = None
        numbers = []

        for ws in workspaces:
            numbers.append(str(ws.num))
            if ws.focused:
                ws_num = ws.num

        non_empty = []
        if self.settings["show-name"] or self.settings["show-icon"]:
            f = self.i3.get_tree().find_focused()
            if f.type == "con" and f.name and str(f.parent.workspace().num) in self.settings["numbers"]:
                win_name = f.name[:self.settings["name-length"]]

                if f.app_id:
                    win_id = f.app_id
                elif f.window_class:
                    win_id = f.window_class

            for item in tree.descendants():
                if item.type == "workspace":
                    # find non-empty workspaces
                    # if self.settings["mark-content"] or self.settings["hide-empty"]:
                    tasks_num = 0
                    for d in item.descendants():
                        if d.type == "con" and d.name:
                            tasks_num += 1
                    if tasks_num > 0:
                        non_empty.append(item.num)

                    for node in item.floating_nodes:
                        if str(node.workspace().num) in self.settings["numbers"]:
                            if node.focused and node.name:
                                win_name = node.name[:self.settings["name-length"] - 1]

                                if node.app_id:
                                    win_id = node.app_id
                                elif node.window_class:
                                    win_id = node.window_class
                                layout = "floating"

                            non_empty.append(node.workspace().num)

            if not layout:
                layout = f.parent.layout

        return ws_num, win_name, win_id, non_empty, layout, numbers

    def on_click(self, event_box, event_button, num):
        nwg_panel.common.i3.command("workspace number {}".format(num))

    def on_scroll(self, event_box, event):
        if event.direction == Gdk.ScrollDirection.UP:
            nwg_panel.common.i3.command("workspace prev")
        elif event.direction == Gdk.ScrollDirection.DOWN:
            nwg_panel.common.i3.command("workspace next")

    def on_enter_notify_event(self, widget, event):
        widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
        widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)

    def on_leave_notify_event(self, widget, event):
        widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
        widget.unset_state_flags(Gtk.StateFlags.SELECTED)



import typing
import multiprocessing as mp

from . import host, watcher
from .tray import Tray

watcher_process = None


def init_tray(trays: typing.List[Tray]):
    # Run host in GLib main loop
    host.init(0, trays)

    # Playerctl and tray watcher are both dbus-driven modules. Some bad media
    # players will freeze on start, because status notifier registration and
    # playerctl pulling metadata happen at the same time. Run watcher in a
    # separate process workarounds this issue.
    global watcher_process
    ctx = mp.get_context('spawn')
    watcher_process = ctx.Process(target=watcher.init, daemon=True)
    watcher_process.start()

def deinit_tray():
    global watcher_process
    if watcher_process:
        watcher_process.terminate()



#!/usr/bin/env python3

"""
This command has two purposes:

1. Execute commands from ~./config/nwg-panel/autostart-dwl.sh file (if found);
2. save data provided by dwl (title, tags, layout for each output) to the cache file
   in json format, for further use by the dwl module.

You need to start dwl with `dwl -s nwg-dwl-interface`.
"""

import subprocess
import fileinput
import os
import sys
import json
from time import sleep


def is_command(cmd):
    cmd = cmd.split()[0]  # strip arguments
    cmd = "command -v {}".format(cmd)
    try:
        is_cmd = subprocess.check_output(cmd, shell=True).decode("utf-8").strip()
        if is_cmd:
            return True

    except subprocess.CalledProcessError:
        return False


def list_outputs():
    print("Checking outputs...")
    outputs = []
    if is_command("wlr-randr"):
        lines = subprocess.check_output("wlr-randr", shell=True).decode("utf-8").strip().splitlines()
        for line in lines:
            if not line.startswith(" "):
                name = line.split()[0]
                print(name)
                outputs.append(name)
    else:
        print("Missing wlr-randr dependency")
    return outputs


def get_cache_dir():
    if os.getenv("XDG_CACHE_HOME"):
        return os.getenv("XDG_CACHE_HOME")
    elif os.getenv("HOME") and os.path.isdir(os.path.join(os.getenv("HOME"), ".cache")):
        return os.path.join(os.getenv("HOME"), ".cache")
    else:
        return None


def get_config_dir():
    xdg_config_home = os.getenv('XDG_CONFIG_HOME')
    config_home = xdg_config_home if xdg_config_home else os.path.join(os.getenv("HOME"), ".config")
    config_dir = os.path.join(config_home, "nwg-panel")
    if os.path.isdir(config_dir):
        return config_dir
    else:
        return None


def main():
    refresh_signal = os.getenv("SIG") if os.getenv("SIG") else 10

    outputs = list_outputs()
    if len(outputs) > 0:
        num_lines = len(outputs) * 4
    else:
        print("No output detected, terminating")
        sys.exit(1)

    data = {}

    # Determine output file location
    cache_dir = get_cache_dir()
    if not cache_dir:
        print("Couldn't detect cache directory")
        sys.exit(1)
    else:
        output_file = os.path.join(cache_dir, "nwg-dwl-data")

    # Run autostart script if present
    config_dir = get_config_dir()
    if not config_dir:
        print("Couldn't detect nwg-panel config directory")
    else:
        autostart = os.path.join(config_dir, "autostart-dwl.sh")
        if os.path.isfile(autostart):
            print("Running {}".format(autostart))
            os.system(autostart)

        sleep(1)

    # remove stale data file, if any
    if os.path.isfile(output_file):
        os.remove(output_file)

    # read stdin, parse data, save in json format
    cnt = 0
    print("num_lines = {}".format(num_lines))
    for line in fileinput.input():
        parts = line.split()

        output = parts[0]
        if output not in data:
            data[output] = {}

        if parts[1] == "title":
            if len(parts) >= 3:
                title = ' '.join(parts[2:])
                if "title" not in data[output] or data[output]["title"] != title:
                    data[output]["title"] = title
            else:
                if "title" not in data[output] or data[output]["title"] != "":
                    data[output]["title"] = ""

        elif parts[1] == "selmon":
            if "selmon" not in data[output] or data[output]["selmon"] != parts[2]:
                data[output]["selmon"] = parts[2]

        elif parts[1] == "tags":
            tags = line.split("{} tags".format(output))[1].strip()
            if "tags" not in data[output] or data[output]["tags"] != tags:
                data[output]["tags"] = tags

        elif parts[1] == "layout":
            if "layout" not in data[output] or data[output]["layout"] != parts[2]:
                data[output]["layout"] = parts[2]

        cnt += 1

        if cnt == num_lines:
            with open(output_file, 'w') as fp:
                json.dump(data, fp, indent=4)

            subprocess.Popen("pkill -f -{} nwg-panel".format(refresh_signal), shell=True)
            cnt = 0


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from nwg_panel.tools import check_key
import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')

from gi.repository import Gtk


class DwlTags(Gtk.EventBox):
    def __init__(self, output, settings):
        Gtk.EventBox.__init__(self)
        check_key(settings, "tag-names", "1 2 3 4 5 6 7 8 9")
        check_key(settings, "title-limit", 55)
        check_key(self.settings, "angle", 0.0)

        self.output = output
        self.settings = settings

        names = self.settings["tag-names"].split()
        self.tags = names if len(names) == 9 else ["1", "2", "3", "4", "5", "6", "7", "8", "9"]

        self.byte_dict = {1: 1, 2: 2, 3: 4, 4: 8, 5: 16, 6: 32, 7: 64, 8: 128, 9: 256}

        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        if self.settings["angle"] != 0.0:
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
        self.add(self.box)
        self.label = Gtk.Label()
        if self.settings["angle"] != 0.0:
            self.label.set_angle(self.settings["angle"])
        self.tag_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5)
        if self.settings["angle"] != 0.0:
            self.tag_box.set_orientation(Gtk.Orientation.VERTICAL)
        self.box.pack_end(self.label, False, False, 4)
        self.show_all()

    def refresh(self, dwl_data):
        if dwl_data:
            try:
                data = dwl_data[self.output]
                tags_string = data["tags"]
                tags = tags_string.split()
                non_empty_output_tags = int(tags[0])
                selected_output_tag = int(tags[1])
                current_win_on_output_tags = int(tags[2])
                urgent_tags = int(tags[3])

                if self.tag_box:
                    self.tag_box.destroy()
                    self.tag_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
                    if self.settings["angle"] != 0.0:
                        self.tag_box.set_orientation(Gtk.Orientation.VERTICAL)
                    self.tag_box.set_property('name', 'dwl-tag-box')
                    self.box.pack_start(self.tag_box, False, False, 4)

                cnt = 1
                win_on_tags = []
                for item in self.tags:
                    tag_wrapper = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
                    if self.settings["angle"] != 0.0:
                        tag_wrapper.set_orientation(Gtk.Orientation.VERTICAL)
                    label = Gtk.Label()
                    if self.settings["angle"] != 0.0:
                        label.set_angle(self.settings["angle"])
                    tag_wrapper.pack_start(label, False, False, 0)
                    if self.byte_dict[cnt] == selected_output_tag:
                        tag_wrapper.set_property('name', "dwl-tag-selected")
                    self.tag_box.pack_start(tag_wrapper, False, False, 1)
                    label.set_text(item)

                    if self.byte_dict[cnt] & non_empty_output_tags != 0:
                        label.set_property('name', "dwl-tag-occupied")
                    else:
                        label.set_property('name', "dwl-tag-free")

                    if self.byte_dict[cnt] & urgent_tags != 0:
                        label.set_property('name', "dwl-tag-urgent")

                    if self.byte_dict[cnt] & current_win_on_output_tags != 0:
                        win_on_tags.append(str(cnt))

                    cnt += 1
                self.tag_box.show_all()

                layout = data["layout"]

                title = data["title"]
                if len(title) > self.settings["title-limit"]:
                    title = title[:self.settings["title-limit"] - 1]

                # title suffix to add if win present on more than 1 tag
                s = ", ".join(win_on_tags) if len(win_on_tags) > 1 else ""
                if s:
                    title = "{} ({})".format(title, s)

                # selmon = data["selmon"] == "1"
                # print("{} {} {}".format(tags_string, layout, title))
                self.label.set_text("{} {}".format(layout, title))
            except KeyError:
                print("No data found for output {}".format(self.output))



#!/usr/bin/env python3

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
gi.require_version('GtkLayerShell', '0.1')
from gi.repository import Gtk, Gdk, GLib, GtkLayerShell

from nwg_panel.tools import check_key, get_brightness, set_brightness, update_image, create_background_task, eprint


class BrightnessSlider(Gtk.EventBox):
    def __init__(self, settings, icons_path=""):
        Gtk.EventBox.__init__(self)

        defaults = {
            "show-values": True,
            "icon-size": 16,
            "interval": 10,
            "hover-opens": False,
            "leave-closes": False,
            "root-css-name": "brightness-module",
            "css-name": "brightness-popup",
            "angle": 0.0,
            "icon-placement": "start",
            "backlight-device": "",
            "backlight-controller": "brightnessctl",
            "slider-orientation": "horizontal",
            "slider-inverted": False,
            "popup-icon-placement": "start",
            "popup-horizontal-alignment": "left",
            "popup-vertical-alignment": "top",
            "popup-width": 256,
            "popup-height": 64,
            "popup-horizontal-margin": 0,
            "popup-vertical-margin": 0,
            "step-size": 1,
        }
        for key in defaults:
            check_key(settings, key, defaults[key])
        self.settings = settings

        self.set_property("name", self.settings["root-css-name"])

        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.add(self.box)

        self.icons_path = icons_path
        self.icon_size = settings["icon-size"]
        self.bri_icon_name = "view-refresh-symbolic"
        self.bri_image = Gtk.Image.new_from_icon_name(self.bri_icon_name, Gtk.IconSize.MENU)

        self.bri_label = Gtk.Label() if settings["show-values"] else None
        self.bri_value = 0

        self.popup_window = PopupWindow(self, settings, icons_path=self.icons_path)

        if settings["angle"] != 0.0:
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
            self.bri_label.set_angle(settings["angle"])

        # events
        self.connect('button-release-event', self.on_button_release)
        self.connect('enter-notify-event', self.on_enter_notify_event)
        self.connect('leave-notify-event', self.on_leave_notify_event)
        if self.settings["step-size"] > 0:
            self.add_events(Gdk.EventMask.SCROLL_MASK) 
            self.connect('scroll-event', self.on_scroll)

        self.build_box()

        self.refresh()

    def build_box(self):
        if self.settings["icon-placement"] == "start":
            self.box.pack_start(self.bri_image, False, False, 2)
            
        if self.bri_label:
            self.box.pack_start(self.bri_label, False, False, 2)

        if self.settings["icon-placement"] == "end":
            self.box.pack_start(self.bri_image, False, False, 2)

    def refresh(self):
        thread = create_background_task(self.refresh_output, self.settings["interval"])
        thread.start()

    def refresh_output(self):
        try:
            self.bri_value = get_brightness(device=self.settings["backlight-device"],
                                            controller=self.settings["backlight-controller"])
            GLib.idle_add(self.update_brightness)
        except Exception as e:
            eprint(e)

    def update_brightness(self, get=True):
        icon_name = bri_icon_name(self.bri_value)

        if icon_name != self.bri_icon_name:
            update_image(self.bri_image, icon_name, self.icon_size, self.icons_path)
            self.bri_icon_name = icon_name

        if self.bri_label:
            self.bri_label.set_text("{}%".format(self.bri_value))
        
        if get:
            self.popup_window.refresh()
    
    def on_button_release(self, w, event):
        if not self.popup_window.get_visible():
            self.popup_window.show_all()
        else:
            self.popup_window.hide()
        
        return False
    
    def on_scroll(self, w, event):
        if event.direction == Gdk.ScrollDirection.UP:
            self.bri_value += self.settings["step-size"]
        elif event.direction == Gdk.ScrollDirection.DOWN:
            self.bri_value -= self.settings["step-size"]
        
        self.bri_value = min(self.bri_value, 100)
        self.bri_value = max(self.bri_value, 0)

        self.update_brightness(get=False)

        if self.popup_window.get_visible():
            self.popup_window.bri_scale.set_value(self.bri_value)

        set_brightness(self.bri_value, device=self.settings["backlight-device"],
                       controller=self.settings["backlight-controller"])

        return False

    def on_enter_notify_event(self, widget, event):
        if self.settings["hover-opens"]:
            if not self.popup_window.get_visible():
                self.popup_window.show_all()
        else:
            widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
            widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)

        # cancel popup window close, as it's probably unwanted ATM
        self.popup_window.on_window_enter()

        return True

    def on_leave_notify_event(self, widget, event):
        widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
        widget.unset_state_flags(Gtk.StateFlags.SELECTED)
        return True


class PopupWindow(Gtk.Window):
    def __init__(self, parent, settings, monitor=None, icons_path=""):
        Gtk.Window.__init__(self, type_hint=Gdk.WindowTypeHint.NORMAL)
        GtkLayerShell.init_for_window(self)
        if monitor:
            GtkLayerShell.set_monitor(self, monitor)

        self.parent = parent
        self.settings = settings
        self.icon_size = settings["icon-size"]
        self.icons_path = icons_path
        self.src_tag = 0
        self.value_changed = False
        self.scrolled = False

        self.set_property("name", self.settings["css-name"])
        
        self.connect("show", self.on_window_show)
        if settings["leave-closes"]:
            self.connect("leave_notify_event", self.on_window_exit)
            self.connect("enter_notify_event", self.on_window_enter)

        eb = Gtk.EventBox()
        eb.set_above_child(False)

        self.box = Gtk.Box(spacing=0)
        if self.settings["slider-orientation"] == "vertical":
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
            self.bri_scale = Gtk.Scale.new_with_range(orientation=Gtk.Orientation.VERTICAL, min=0, max=100, step=1)
        elif self.settings["slider-orientation"] == "horizontal":
            self.box.set_orientation(Gtk.Orientation.HORIZONTAL)
            self.bri_scale = Gtk.Scale.new_with_range(orientation=Gtk.Orientation.HORIZONTAL, min=0, max=100, step=1)
        self.bri_scale.set_inverted(self.settings["slider-inverted"])

        if self.settings["backlight-controller"] == "ddcutil":
            self.bri_scale_handler = self.bri_scale.connect("value-changed", self.on_value_changed)
            self.bri_scale.connect("button-release-event", self.on_button_release)
            
            self.bri_scale.add_events(Gdk.EventMask.SCROLL_MASK)
            self.bri_scale.connect('scroll-event', self.on_scroll)
        else:
            self.bri_scale_handler = self.bri_scale.connect("value-changed", self.set_bri)

        self.bri_icon_name = "view-refresh-symbolic"
        self.bri_image = Gtk.Image.new_from_icon_name(self.bri_icon_name, Gtk.IconSize.MENU)
            
        eb.add(self.box)
        self.add(eb)

        if settings["popup-vertical-alignment"] == "top":
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.TOP, True)
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.TOP, settings["popup-vertical-margin"])
        elif settings["popup-vertical-alignment"] == "bottom":
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.BOTTOM, True)
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.BOTTOM, settings["popup-vertical-margin"])

        if settings["popup-horizontal-alignment"] == "left":
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.LEFT, True)
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.LEFT, settings["popup-horizontal-margin"])
        elif settings["popup-horizontal-alignment"] == "right":
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.RIGHT, True)
            GtkLayerShell.set_margin(self, GtkLayerShell.Edge.RIGHT, settings["popup-horizontal-margin"])

        Gtk.Widget.set_size_request(self.box, settings["popup-width"], settings["popup-height"])

        self.build_box()

        Gdk.threads_add_timeout(GLib.PRIORITY_LOW, 500, self.refresh)
    
    def build_box(self):
        if self.settings["popup-icon-placement"] == "start":
            self.box.pack_start(self.bri_image, False, False, 6)
        
        self.box.pack_start(self.bri_scale, True, True, 5)
        
        if self.settings["popup-icon-placement"] == "end":
            self.box.pack_start(self.bri_image, False, False, 6)

    def refresh(self, *args):
        if self.get_visible():
            if not self.value_changed:
                self.bri_scale.set_value(self.parent.bri_value)
            if self.parent.bri_icon_name != self.bri_icon_name:
                update_image(self.bri_image, self.parent.bri_icon_name, self.icon_size, self.icons_path)
                self.bri_icon_name = self.parent.bri_icon_name

        else:
            with self.bri_scale.handler_block(self.bri_scale_handler):
                self.bri_scale.set_value(self.parent.bri_value)

        return True

    def on_window_exit(self, w, e):
        if self.get_visible():
            self.src_tag = GLib.timeout_add_seconds(1, self.hide_and_clear_tag)
        return True

    def hide_and_clear_tag(self):
        self.hide()
        self.src_tag = 0

    def on_window_enter(self, *args):
        if self.src_tag > 0:
            GLib.Source.remove(self.src_tag)
            self.src_tag = 0
        return True

    def on_window_show(self, *args):
        self.src_tag = 0
        self.refresh()

    def set_bri(self, slider):
        self.parent.bri_value = int(slider.get_value())
        self.parent.update_brightness(get=False)
        set_brightness(self.parent.bri_value, device=self.settings["backlight-device"],
                       controller=self.settings["backlight-controller"])

    def on_button_release(self, scale, event):
        if self.value_changed:
            self.set_bri(scale)
            self.value_changed = False

    def on_value_changed(self, *args):
        if self.scrolled:
            self.set_bri(self.bri_scale)
            self.scrolled = False
        else:
            self.value_changed = True

    def on_scroll(self, w, event):
        self.scrolled = True

def bri_icon_name(value):
    icon_name = "display-brightness-low-symbolic"
    if value > 70:
        icon_name = "display-brightness-high-symbolic"
    elif value > 30:
        icon_name = "display-brightness-medium-symbolic"

    return icon_name



try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

try:
    __version__ = metadata.version("nwg-panel")
except Exception:
    __version__ = "unknown"



#!/usr/bin/env python3

from gi.repository import GLib

import psutil

import gi

from nwg_panel.tools import create_background_task

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')

from gi.repository import Gtk, Gdk


class CpuAvg(Gtk.EventBox):
    def __init__(self):
        self.avg = 0
        self.cnt = 0
        Gtk.EventBox.__init__(self)
        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.add(self.box)
        self.label = Gtk.Label()
        self.label.set_property("name", "executor-label")

        self.build_box()
        self.refresh()

    def update_widget(self, val, cnt):
        self.label.set_text(val)
        self.label.set_tooltip_text("{} checks".format(cnt))

        return False

    def get_output(self):
        try:
            val = psutil.cpu_percent(interval=1)
            self.avg = self.avg + val
            self.cnt += 1
            val = "{:.2f}%".format(round(self.avg / self.cnt, 2))
            GLib.idle_add(self.update_widget, val, str(self.cnt))
        except Exception as e:
            print(e)

    def refresh(self):
        thread = create_background_task(self.get_output, 2)
        thread.start()

    def build_box(self):
        self.box.pack_start(self.label, False, False, 4)
        self.label.show()



#!/usr/bin/env python3
import json
import os

import gi

from nwg_panel.tools import check_key, update_image, create_background_task, eprint, hyprctl, niri_keyboard_layouts, \
    niri_ipc

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')

from gi.repository import Gtk, Gdk, GLib


def on_enter_notify_event(widget, event):
    widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
    widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)


def on_leave_notify_event(widget, event, *args):
    widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
    widget.unset_state_flags(Gtk.StateFlags.SELECTED)


def on_menu_popped_up(arg0, arg1, arg2, arg3, arg4, widget):
    widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
    widget.unset_state_flags(Gtk.StateFlags.SELECTED)


class KeyboardLayout(Gtk.EventBox):
    def __init__(self, settings, icons_path):
        self.settings = settings
        self.icons_path = icons_path
        Gtk.EventBox.__init__(self)
        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.add(self.box)
        self.image = Gtk.Image()
        self.label = Gtk.Label.new("")
        self.icon_path = None

        if os.getenv("SWAYSOCK"):
            from i3ipc import Connection
            self.i3 = Connection()
            self.compositor = "sway"
        elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
            self.compositor = "Hyprland"
        elif os.getenv("NIRI_SOCKET"):
            self.compositor = "niri"
        else:
            self.compositor = ""
            eprint("KeyboardLayout module only supports sway, Hyprland or Niri")

        if self.compositor:
            self.keyboards = self.list_keyboards()
            if self.keyboards or self.compositor == "niri":
                self.keyboard_names = []
                for k in self.keyboards:
                    if self.compositor == "Hyprland":
                        self.keyboard_names.append(k["name"])
                    # On sway some devices may be listed twice, let's add them just once
                    elif k.identifier not in self.keyboard_names:
                        self.keyboard_names.append(k.identifier)

                self.kb_layouts = self.get_kb_layouts()

                check_key(settings, "keyboard-device-sway", "")
                check_key(settings, "keyboard-device-hyprland", "")
                self.device_name = settings["keyboard-device-sway"] if self.compositor == "sway" else settings[
                    "keyboard-device-hyprland"]

                check_key(settings, "root-css-name", "root-executor")
                check_key(settings, "css-name", "executor-label")
                check_key(settings, "icon-placement", "left")
                check_key(settings, "icon-size", 16)
                check_key(settings, "show-icon", True)
                check_key(settings, "interval", 0)
                check_key(settings, "tooltip-text", "LMB: Next layout, RMB: Menu")
                check_key(settings, "angle", 0.0)

                self.label.set_angle(settings["angle"])

                if settings["angle"] != 0.0:
                    self.box.set_orientation(Gtk.Orientation.VERTICAL)

                update_image(self.image, "input-keyboard", self.settings["icon-size"], self.icons_path)

                self.set_property("name", settings["root-css-name"])
                if settings["css-name"]:
                    self.label.set_property("name", settings["css-name"])
                else:
                    self.label.set_property("name", "executor-label")

                if settings["tooltip-text"]:
                    self.set_tooltip_text(settings["tooltip-text"])

                self.connect('button-release-event', self.on_button_release)
                self.connect('enter-notify-event', on_enter_notify_event)
                self.connect('leave-notify-event', on_leave_notify_event)

                self.build_box()
                label = self.get_current_layout()
                if label:
                    self.label.set_text(label)
                self.show_all()
            else:
                print("KeyboardLayout module: failed listing devices, won't create UI, sorry.")

        self.refresh()

    def list_keyboards(self):
        if self.compositor == "Hyprland":
            o = hyprctl("j/devices")
            devices = json.loads(o)
            keyboards = devices["keyboards"] if "keyboards" in devices else []
        elif self.compositor == "sway":
            inputs = self.i3.get_inputs()
            keyboards = []
            for i in inputs:
                if i.type == "keyboard":
                    keyboards.append(i)
        else:
            keyboards = []

        return keyboards

    def get_kb_layouts(self):
        if self.compositor == "Hyprland":
            o = hyprctl("j/getoption input:kb_layout")
            option = json.loads(o)
            if option and "str" in option:
                return option["str"].split(",")
            return []
        elif self.compositor == "sway":
            layout_names = []
            if self.keyboards:
                for k in self.keyboards:
                    for name in k.xkb_layout_names:
                        if name not in layout_names:
                            layout_names.append(name)
            return layout_names
        elif self.compositor == "niri":
            return niri_keyboard_layouts()["names"]
        else:
            return []

    def get_current_layout(self):
        if self.compositor == "Hyprland":
            if self.device_name:
                for k in self.keyboards:
                    if k["name"] == self.device_name:
                        return k["active_keymap"]
                return "unknown"
            else:
                for k in self.keyboards:
                    if "keyboard" in k["name"]:
                        return k["active_keymap"]
                return self.keyboards[0]["layout"]
        elif self.compositor == "sway":
            for k in self.keyboards:
                if "keyboard" in k.identifier:
                    return k.xkb_active_layout_name
                return self.keyboards[0].xkb_active_layout_name
            return "unknown"
        elif self.compositor == "niri":
            nkl = niri_keyboard_layouts()
            return nkl["names"][nkl["current_idx"]]
        else:
            return "unknown"

    def update_label(self):
        self.keyboards = self.list_keyboards()
        txt = self.get_current_layout()
        if txt:
            self.label.set_text(txt)

    def refresh(self, *args):
        thread = create_background_task(self.update_label, self.settings["interval"])
        thread.start()

    def build_box(self):
        if self.settings["show-icon"] and self.settings["icon-placement"] == "left":
            self.box.pack_start(self.image, False, False, 3)
        self.box.pack_start(self.label, False, False, 3)
        if self.settings["show-icon"] and self.settings["icon-placement"] != "left":
            self.box.pack_start(self.image, False, False, 3)

    def on_left_click(self):
        if self.compositor == "Hyprland":
            if self.device_name:
                # apply to selected device
                hyprctl(f"switchxkblayout {self.device_name} next")
            else:
                # apply to all devices
                for name in self.keyboard_names:
                    hyprctl(f"switchxkblayout {name} next")
        elif self.compositor == "sway":
            # apply to all devices of type:keyboard
            self.i3.command(f'input type:keyboard xkb_switch_layout next')
        elif self.compositor == "niri":
            command = {"Action":{"SwitchLayout":{"layout":"Next"}}}
            niri_ipc(json.dumps(command), is_json=True)

        self.update_label()

    def on_menu_item(self, item, idx):
        if self.compositor == "Hyprland":
            if self.device_name:
                # apply to selected device
                hyprctl(f'switchxkblayout {self.device_name} {idx}')
            else:
                # apply to all devices
                for name in self.keyboard_names:
                    hyprctl(f'switchxkblayout {name} {idx}')
        elif self.compositor == "sway":
            # apply to all devices of type:keyboard
            self.i3.command(f'input type:keyboard xkb_switch_layout {idx}')
        elif self.compositor == "niri":
            command = {"Action":{"SwitchLayout":{"layout":{"Index":idx}}}}
            niri_ipc(json.dumps(command), is_json=True)

        self.update_label()

    def on_right_click(self):
        menu = Gtk.Menu()
        menu.connect("popped-up", on_menu_popped_up, self)
        for i in range(len(self.kb_layouts)):
            item = Gtk.MenuItem.new_with_label(self.kb_layouts[i])
            item.connect("activate", self.on_menu_item, i)
            menu.append(item)
        menu.set_reserve_toggle_size(False)
        menu.show_all()
        menu.popup_at_widget(self.label, Gdk.Gravity.STATIC, Gdk.Gravity.STATIC, None)


    def on_button_release(self, widget, event):
        if event.button == 1:
            self.on_left_click()
        elif event.button == 3:
            self.on_right_click()



#!/usr/bin/env python3

import json
import subprocess

from gi.repository import Gtk, GLib
from i3ipc import Event

from nwg_panel.tools import check_key, update_image


class SwayMode(Gtk.Box):
    def __init__(self, i3, settings, icons_path=""):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.settings = settings
        self.i3 = i3
        self.label = Gtk.Label.new("default")
        self.mode = "default"

        icon = Gtk.Image()

        defaults = {
            "show-default": False,
            "show-icon": True,
            "css-name": "executor-label",
            "icon-size": 16,
            "angle": 0.0
        }
        for key in defaults:
            check_key(settings, key, defaults[key])

        update_image(icon, "object-flip-vertical", self.settings["icon-size"], icons_path)

        if self.settings["css-name"]:
            self.set_property("name", self.settings["css-name"])
        else:
            self.set_property("name", "executor-label")

        if settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)
            self.label.set_angle(settings["angle"])

        if settings["show-icon"]:
            self.pack_start(icon, False, False, 2)

        self.pack_start(self.label, False, False, 2)

        self.check_initial_mode()

        if self.mode == "default" and not self.settings["show-default"]:
            GLib.idle_add(self.hide, priority=GLib.PRIORITY_HIGH)

        self.i3.on(Event.MODE, self.on_i3ipc_event)

    def check_initial_mode(self):
        # On panel startup we may already be in some mode other than default,
        # but we only check it on i3ipc event.
        # As i3ipc seemingly does not implement the `get_binding_state` function,
        # for initial check we need to parse `swaymsg -t get_binding_state` output.
        self.mode = "default"
        o = subprocess.check_output("swaymsg -t get_binding_state".split()).decode("utf-8")
        try:
            self.mode = json.loads(o)["name"]
        except KeyError:
            pass
        self.set_visibility()

    def on_i3ipc_event(self, i3conn, event):
        self.mode = event.change
        self.set_visibility()

    def set_visibility(self):
        self.label.set_text(self.mode)
        if self.mode != "default" or self.settings["show-default"]:
            self.show_all()
        else:
            self.hide()



#!/usr/bin/env python3

import json

from gi.repository import Gtk, Gdk

from nwg_panel.tools import hyprctl, update_image, update_image_fallback_desktop, eprint


class HyprlandTaskbar(Gtk.Box):
    def __init__(self, settings, position, monitors, workspaces, clients, activewindow, display_name="", icons_path=""):
        defaults = {
            "name-max-len": 24,
            "image-size": 16,
            "workspaces-spacing": 0,
            "client-padding": 0,
            "show-app-icon": True,
            "show-app-name": True,
            "show-app-name-special": False,
            "show-layout": True,
            "all-outputs": False,
            "all-workspaces": True,
            "show-ws-names": True,
            "mark-xwayland": True,
            "angle": 0.0
        }
        for key in defaults:
            if key not in settings:
                settings[key] = defaults[key]
        self.settings = settings

        self.position = position
        self.display_name = display_name
        self.icons_path = icons_path

        self.monitors = None
        self.mon_id2name = {}
        self.active_workspaces = None
        self.clients = None
        self.ws_nums = None
        self.workspaces = None
        self.activewindow = None
        # default values, in case we find no workspaces defined in workspacerules
        self.ws_strings = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]

        # replace self.ws_strings with workspaces defined in workspacerules, if any
        ws_strings = []
        o = hyprctl("j/workspacerules")
        ws_dicts = json.loads(o)
        if len(ws_dicts) > 0:
            for item in json.loads(o):  # list of dicts
                ws_strings.append(item["workspaceString"])
            self.ws_strings = ws_strings

        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=settings["workspaces-spacing"])
        if self.settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)

        self.refresh(monitors, workspaces, clients, activewindow)

    def parse_monitors(self, monitors):
        self.monitors = monitors
        self.active_workspaces = []
        for m in self.monitors:
            self.mon_id2name[m["id"]] = m["name"]
            self.active_workspaces.append(m["activeWorkspace"]["id"])

    def parse_workspaces(self, ws):
        self.ws_nums = []
        self.workspaces = {}
        for item in ws:
            self.ws_nums.append(item["id"])
            self.workspaces[item["id"]] = item
        self.ws_nums.sort()

    def refresh(self, monitors, workspaces, clients, activewindow):
        self.parse_monitors(monitors)
        self.parse_workspaces(workspaces)
        # Turned off due to https://github.com/hyprwm/Hyprland/issues/2413
        # self.parse_clients(clients)
        self.clients = clients
        self.activewindow = activewindow
        for item in self.get_children():
            item.destroy()
        self.build_box()

    def build_box(self):
        for ws_num in self.ws_nums:
            ws_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
            if self.settings["angle"] != 0.0:
                ws_box.set_orientation(Gtk.Orientation.VERTICAL)
            self.pack_start(ws_box, False, False, 0)
            if self.workspaces[ws_num]["monitor"] == self.display_name or self.settings["all-outputs"]:
                eb = Gtk.EventBox()

                if not self.settings["all-workspaces"] and ws_num not in self.active_workspaces:
                    continue
                if self.settings["show-ws-names"]:
                    ws_box.pack_start(eb, False, False, 6)
                    lbl = Gtk.Label()
                    if ws_num in self.active_workspaces:
                        lbl.set_markup("<u>{}</u>:".format(self.workspaces[ws_num]["name"]))
                    else:
                        lbl.set_text("{}:".format(self.workspaces[ws_num]["name"]))
                    eb.add(lbl)
                cl_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
                ws_box.pack_start(cl_box, False, False, 0)
                for client in self.clients:
                    # if client["title"] prevents from creation of ghost client boxes
                    if client["title"] and client["workspace"]["id"] == ws_num:
                        client_box = ClientBox(self.settings, client, self.position, self.icons_path, self.ws_strings)
                        if self.activewindow and client["address"] == self.activewindow["address"]:
                            client_box.box.set_property("name", "task-box-focused")
                        else:
                            client_box.box.set_property("name", "task-box")
                        cl_box.pack_start(client_box, False, False, self.settings["client-padding"])

        self.show_all()

    def on_ws_click(self, widget, event, ws_num):
        res = hyprctl("dispatch workspace name:{}".format(ws_num))
        if res != "ok":
            hyprctl(f'dispatch hl.dsp.focus({{ workspace = "{ws_num}"}})')


def on_enter_notify_event(widget, event):
    widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
    widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)


def on_leave_notify_event(widget, event):
    widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
    widget.unset_state_flags(Gtk.StateFlags.SELECTED)


class ClientBox(Gtk.EventBox):
    def __init__(self, settings, client, position, icons_path, ws_strings):
        self.position = position
        self.settings = settings
        self.address = client["address"]
        self.floating = client["floating"]
        self.icons_path = icons_path
        self.ws_strings = ws_strings
        Gtk.EventBox.__init__(self)
        self.box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=0)
        if settings["angle"] != 0.0:
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
        self.add(self.box)

        self.connect('enter-notify-event', on_enter_notify_event)
        self.connect('leave-notify-event', on_leave_notify_event)
        if client["workspace"]["name"] == "special":
            self.connect('button-release-event', self.on_special)
        else:
            self.connect('button-release-event', self.on_click, client, self.box)

        image = None
        if settings["show-app-icon"]:
            name = client["class"]
            image = Gtk.Image()
            image.set_property("name", "task-box-icon")
            update_image_fallback_desktop(image, name, settings["image-size"], icons_path)
            self.box.pack_start(image, False, False, 4)

        name = client["title"][:settings["name-max-len"]]
        if settings["mark-xwayland"] and client["xwayland"]:
            name = "X|" + name

        if settings["show-app-name"]:
            if "special" not in client["workspace"]["name"] or settings["show-app-name-special"]:
                lbl = Gtk.Label()
                lbl.set_angle(self.settings["angle"])

                lbl.set_text(name)
                self.box.pack_start(lbl, False, False, 6)
            else:
                if name and image:
                    image.set_tooltip_text(name)

        if settings["show-layout"]:
            if client["pinned"]:
                img = Gtk.Image()
                update_image(img, "pin", self.settings["image-size"], self.icons_path)
                self.box.pack_start(img, False, False, 0)

            elif client["floating"]:
                img = Gtk.Image()
                update_image(img, "focus-windows", self.settings["image-size"], self.icons_path)
                self.box.pack_start(img, False, False, 0)

    def on_click(self, widget, event, client, popup_at_widget):
        if event.button == 1:
            res = hyprctl(f"dispatch focuswindow address:{self.address}")
            # handle new Lua dispatcher on Hyprland >= v0.55.0
            if res != "ok":
                hyprctl(f'dispatch hl.dsp.focus({{ window = "address:{self.address}" }})')
        if event.button == 3:
            menu = self.context_menu(client)
            menu.show_all()
            if self.position == "bottom":
                menu.popup_at_widget(popup_at_widget, Gdk.Gravity.SOUTH, Gdk.Gravity.NORTH, None)
            else:
                menu.popup_at_widget(popup_at_widget, Gdk.Gravity.NORTH, Gdk.Gravity.SOUTH, None)

    def on_special(self, widget, event):
        res = hyprctl('dispatch togglespecialworkspace ')
        if res != "ok":
            hyprctl('dispatch hl.dsp.workspace.toggle_special')


    def context_menu(self, client):
        menu = Gtk.Menu()
        menu.set_reserve_toggle_size(False)

        # Move to workspace
        for _item in self.ws_strings:
            hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 3)
            hbox.set_property("halign", Gtk.Align.START)
            img = Gtk.Image()
            update_image(img, "go-next", 16, self.icons_path)
            hbox.pack_start(img, True, True, 0)
            lbl = Gtk.Label.new(_item)
            hbox.pack_start(lbl, False, False, 0)
            item = Gtk.MenuItem()
            item.add(hbox)
            item.connect("activate", self.movetoworkspace, _item)
            item.set_tooltip_text(f"movetoworkspace {_item}")
            menu.append(item)

        # Toggle floating
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "view-paged-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        img = Gtk.Image()
        update_image(img, "view-dual-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.toggle_floating)
        item.set_tooltip_text("togglefloating")
        menu.append(item)

        # Fullscreen
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "view-fullscreen", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.fullscreen)
        item.set_tooltip_text("fullscreen")
        menu.append(item)

        # Pin
        if client["floating"]:
            hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
            img = Gtk.Image()
            update_image(img, "pin", 16, self.icons_path)
            hbox.pack_start(img, True, True, 0)
            item = Gtk.MenuItem()
            item.add(hbox)
            item.connect("activate", self.pin)
            item.set_tooltip_text("pin")
            menu.append(item)

        # Close
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "window-close", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.close, self.address)
        item.set_tooltip_text("closewindow")
        menu.append(item)

        return menu

    def close(self, *args):
        res = hyprctl(f"dispatch closewindow address:{self.address}")
        if res != "ok":
            hyprctl(f'dispatch hl.dsp.window.close({{ window = "address:{self.address}" }})')

    def toggle_floating(self, *args):
        res = hyprctl("dispatch togglefloating address:{}".format(self.address))
        if res != "ok":
            hyprctl(f'dispatch hl.dsp.window.float({{ action = "toggle", window = "address:{self.address}"}})')

    def fullscreen(self, *args):
        res = hyprctl("dispatch fullscreen address:{}".format(self.address))
        if res != "ok":
            hyprctl(f'dispatch hl.dsp.window.fullscreen({{ action = "toggle", window = "address:{self.address}"}})')

    def pin(self, *args):
        res = hyprctl("dispatch pin address:{}".format(self.address))
        if res == "ok":
            # The above doesn't trigger any event. We need a workaround:
            hyprctl("dispatch focuswindow title:''")
            hyprctl("dispatch focuswindow address:{}".format(self.address))
        else:
            hyprctl(f'dispatch hl.dsp.window.pin({{ window = "address:{self.address}" }})')

    def movetoworkspace(self, menuitem, ws_num):
        res = hyprctl("dispatch movetoworkspace {},address:{}".format(ws_num, self.address))
        if res != "ok":
            hyprctl(f'dispatch hl.dsp.window.move({{ workspace = "{ws_num}", window = "address:{self.address}"}})')



import typing

from dasbus.connection import SessionMessageBus
from dasbus.loop import EventLoop
from dasbus.signal import Signal
from dasbus.client.observer import DBusObserver
from dasbus.server.interface import accepts_additional_arguments
import dasbus.typing

WATCHER_SERVICE_NAME = "org.kde.StatusNotifierWatcher"
WATCHER_OBJECT_PATH = "/StatusNotifierWatcher"

dasbus_event_loop: typing.Union[EventLoop, None] = None


class StatusNotifierWatcherInterface(object):
    __dbus_xml__ = """
        <!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN" "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
        <node>
            <interface name="org.kde.StatusNotifierWatcher">
                <annotation name="org.gtk.GDBus.C.Name" value="Watcher" />

                <method name="RegisterStatusNotifierItem">
                    <annotation name="org.gtk.GDBus.C.Name" value="RegisterItem" />
                    <arg name="service" type="s" direction="in"/>
                </method>

                <method name="RegisterStatusNotifierHost">
                    <annotation name="org.gtk.GDBus.C.Name" value="RegisterHost" />
                    <arg name="service" type="s" direction="in"/>
                </method>

                <property name="RegisteredStatusNotifierItems" type="as" access="read">
                    <annotation name="org.gtk.GDBus.C.Name" value="RegisteredItems" />
                    <annotation name="org.qtproject.QtDBus.QtTypeName.Out0" value="QStringList"/>
                </property>

                <property name="IsStatusNotifierHostRegistered" type="b" access="read">
                    <annotation name="org.gtk.GDBus.C.Name" value="IsHostRegistered" />
                </property>

                <property name="ProtocolVersion" type="i" access="read"/>

                <signal name="StatusNotifierItemRegistered">
                    <annotation name="org.gtk.GDBus.C.Name" value="ItemRegistered" />
                    <arg type="s" direction="out" name="service" />
                </signal>

                <signal name="StatusNotifierItemUnregistered">
                    <annotation name="org.gtk.GDBus.C.Name" value="ItemUnregistered" />
                    <arg type="s" direction="out" name="service" />
                </signal>

                <signal name="StatusNotifierHostRegistered">
                    <annotation name="org.gtk.GDBus.C.Name" value="HostRegistered" />
                </signal>

                <signal name="StatusNotifierHostUnregistered">
                    <annotation name="org.gtk.GDBus.C.Name" value="HostUnregistered" />
                </signal>

            </interface>
        </node>
    """

    PropertiesChanged = Signal()
    StatusNotifierItemRegistered = Signal()
    StatusNotifierItemUnregistered = Signal()
    StatusNotifierHostRegistered = Signal()
    StatusNotifierHostUnregistered = Signal()

    def __init__(self):
        self._statusNotifierItems = []
        self._statusNotifierHosts = []
        self._isStatusNotifierHostRegistered = False
        self._protocolVersion = 0
        self.session_bus = SessionMessageBus()

    def __del__(self):
        self.session_bus.disconnect()

    @accepts_additional_arguments
    def RegisterStatusNotifierItem(self, service, call_info):
        """print(
            "StatusNotifierWatcher -> RegisterStatusNotifierItem\n  service: {}\n  sender: {}".format(
                service,
                call_info["sender"]
            )
        )"""

        # libappindicator sends object path, use sender name and object path
        if service[0] == "/":
            full_service_name = "{}{}".format(call_info["sender"], service)

        # xembedsniproxy sends item name, use the item from the argument
        elif service[0] == ":":
            full_service_name = "{}{}".format(service, "/StatusNotifierItem")

        else:
            full_service_name = "{}{}".format(call_info["sender"], "/StatusNotifierItem")

        if full_service_name not in self._statusNotifierItems:
            item_service_observer = DBusObserver(
                message_bus=self.session_bus,
                service_name=call_info["sender"]
            )
            item_service_observer.service_available.connect(
                lambda _observer: self.item_available_handler(full_service_name)
            )
            item_service_observer.service_unavailable.connect(
                lambda _observer: self.item_unavailable_handler(full_service_name)
            )
            item_service_observer.connect_once_available()
        else:
            """print(
                (
                    "StatusNotifierWatcher -> RegisterStatusNotifierItem: item already registered\n"
                    "  full_service_name: {}"
                ).format(full_service_name, service)
            )"""

    @accepts_additional_arguments
    def RegisterStatusNotifierHost(self, service, call_info):
        # print("StatusNotifierWatcher -> RegisterStatusNotifierHost: {}".format(service))
        if call_info["sender"] not in self._statusNotifierHosts:
            host_service_observer = DBusObserver(
                message_bus=self.session_bus,
                service_name=call_info["sender"]
            )
            host_service_observer.service_available.connect(
                self.host_available_handler
            )
            host_service_observer.service_unavailable.connect(
                self.host_unavailable_handler
            )
            host_service_observer.connect_once_available()
        """else:
            print(
                "StatusNotifierWatcher -> RegisterStatusNotifierHost: host already registered\n  service: {}\n  sender: {})".format(
                    service,
                    call_info["sender"]
                )
            )"""

    @property
    def RegisteredStatusNotifierItems(self) -> list:
        # print("StatusNotifierWatcher -> RegisteredStatusNotifierItems")
        return self._statusNotifierItems

    @property
    def IsStatusNotifierHostRegistered(self) -> bool:
        """print(
            "StatusNotifierWatcher -> IsStatusNotifierHostRegistered: {}".format(
                str(len(self._statusNotifierHosts) > 0)
            )
        )"""
        return len(self._statusNotifierHosts) > 0

    @property
    def ProtocolVersion(self) -> int:
        # print("StatusNotifierWatcher -> ProtocolVersion: ".format(str(self._protocolVersion)))
        return self._protocolVersion

    def item_available_handler(self, full_service_name):
        """print(
            "StatusNotifierWatcher -> item_available_handler\n  full_service_name: {}".format(
                full_service_name
            )
        )"""
        self._statusNotifierItems.append(full_service_name)
        self.StatusNotifierItemRegistered.emit(full_service_name)
        self.PropertiesChanged.emit(WATCHER_SERVICE_NAME, {
            "RegisteredStatusNotifierItems": dasbus.typing.get_variant(
                dasbus.typing.List[dasbus.typing.Str],
                self._statusNotifierItems
            )
        }, [])

    def item_unavailable_handler(self, full_service_name):
        """print(
            "StatusNotifierWatcher -> item_unavailable_handler\n  full_service_name: {}".format(
                full_service_name
            )
        )"""
        if full_service_name in set(self._statusNotifierItems):
            self._statusNotifierItems.remove(full_service_name)
            self.StatusNotifierItemUnregistered.emit(full_service_name)
            self.PropertiesChanged.emit(WATCHER_SERVICE_NAME, {
                "RegisteredStatusNotifierItems": dasbus.typing.get_variant(
                    dasbus.typing.List[dasbus.typing.Str],
                    self._statusNotifierItems
                )
            }, [])

    def host_available_handler(self, observer):
        self._statusNotifierHosts.append(observer.service_name)
        self.StatusNotifierHostRegistered.emit()
        self.PropertiesChanged.emit(WATCHER_SERVICE_NAME, {
            "IsStatusNotifierHostRegistered": dasbus.typing.get_variant(dasbus.typing.Bool, True)
        }, [])

    def host_unavailable_handler(self, observer):
        self._statusNotifierHosts.remove(observer.service_name)
        self.StatusNotifierHostUnregistered.emit()
        if len(self._statusNotifierHosts) == 0:
            self.PropertiesChanged.emit(WATCHER_SERVICE_NAME, {
                "IsStatusNotifierHostRegistered": dasbus.typing.get_variant(dasbus.typing.Bool, False)
            }, [])
        # deinit on host (parent process) unavailable to avoid becoming zombie process
        deinit()


def init():
    session_bus = SessionMessageBus()
    session_bus.publish_object(WATCHER_OBJECT_PATH, StatusNotifierWatcherInterface())
    session_bus.register_service(WATCHER_SERVICE_NAME)
    # print("watcher.init(): published {}{} on dbus.".format(WATCHER_SERVICE_NAME, WATCHER_OBJECT_PATH))

    global dasbus_event_loop
    if dasbus_event_loop is None:
        # print("watcher.init(): running dasbus.EventLoop")
        dasbus_event_loop = EventLoop()
        dasbus_event_loop.run()


def deinit():
    global dasbus_event_loop
    if dasbus_event_loop is not None:
        # print("watcher.deinit(): quitting dasbus.EventLoop")
        dasbus_event_loop.quit()
        dasbus_event_loop = None



#!/usr/bin/env python3

import os
import sys
import json
import subprocess
import stat
import time
import socket
import threading
import re
import glob

import gi

import nwg_panel.common
from nwg_panel.icons import get_icon_name

gi.require_version('GdkPixbuf', '2.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')

from gi.repository import Gtk, Gdk, GdkPixbuf
from shutil import copyfile, which
from datetime import datetime

try:
    import psutil
except:
    pass


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def temp_dir():
    if os.getenv("TMPDIR"):
        return os.getenv("TMPDIR")
    elif os.getenv("TEMP"):
        return os.getenv("TEMP")
    elif os.getenv("TMP"):
        return os.getenv("TMP")

    return "/tmp"


def local_dir():
    local_dir = os.path.join(os.path.join(os.getenv("HOME"), ".local/share/nwg-panel"))
    if not os.path.isdir(local_dir):
        print("Creating '{}'".format(local_dir))
        os.makedirs(local_dir, exist_ok=True)

    return local_dir


def get_config_dir():
    """
    Determine config dir path, create if not found, then create sub-dirs
    :return: config dir path
    """
    xdg_config_home = os.getenv('XDG_CONFIG_HOME')
    config_home = xdg_config_home if xdg_config_home else os.path.join(os.getenv("HOME"), ".config")
    config_dir = os.path.join(config_home, "nwg-panel")
    if not os.path.isdir(config_dir):
        print("Creating '{}'".format(config_dir))
        os.makedirs(config_dir, exist_ok=True)

    # Icon folders to store user-defined icon replacements
    folder = os.path.join(config_dir, "icons_light")
    if not os.path.isdir(folder):
        print("Creating '{}'".format(folder))
        os.makedirs(folder, exist_ok=True)

    folder = os.path.join(config_dir, "icons_dark")
    if not os.path.isdir(os.path.join(folder)):
        print("Creating '{}'".format(folder))
        os.makedirs(folder, exist_ok=True)

    folder = os.path.join(config_dir, "icons_color")
    if not os.path.isdir(os.path.join(folder)):
        print("Creating '{}'".format(folder))
        os.makedirs(folder, exist_ok=True)

    folder = os.path.join(config_dir, "executors")
    if not os.path.isdir(os.path.join(folder)):
        print("Creating '{}'".format(folder))
        os.makedirs(folder, exist_ok=True)

    return config_dir


def copy_files(src_dir, dst_dir, restore=False):
    src_files = os.listdir(src_dir)
    for file in src_files:
        if os.path.isfile(os.path.join(src_dir, file)):
            if not os.path.isfile(os.path.join(dst_dir, file)) or restore:
                copyfile(os.path.join(src_dir, file), os.path.join(dst_dir, file))
                print("Copying '{}'".format(os.path.join(dst_dir, file)))


def copy_executors(src_dir, dst_dir):
    src_files = os.listdir(src_dir)
    for file in src_files:
        if os.path.isfile(os.path.join(src_dir, file)) and not os.path.isfile(os.path.join(dst_dir, file)):
            copyfile(os.path.join(src_dir, file), os.path.join(dst_dir, file))
            print("Copying '{}', marking executable".format(os.path.join(dst_dir, file)))
            st = os.stat(os.path.join(dst_dir, file))
            os.chmod(os.path.join(dst_dir, file), st.st_mode | stat.S_IEXEC)


def load_text_file(path):
    try:
        with open(path, 'r') as file:
            data = file.read()
            return data
    except Exception as e:
        print(e)
        return None


def load_json(path):
    try:
        with open(path, 'r') as f:
            return json.load(f)
    except Exception as e:
        eprint("Error loading json: {}".format(e))
        return {}


def save_json(src_dict, path):
    try:
        with open(path, 'w') as f:
            json.dump(src_dict, f, indent=2)
        return "ok"
    except Exception as e:
        return e


def save_string(string, file):
    try:
        file = open(file, "wt")
        file.write(string)
        file.close()
    except:
        print("Error writing file '{}'".format(file))


def load_string(path):
    try:
        with open(path, 'r') as file:
            data = file.read()
            return data
    except:
        return ""


def load_autotiling():
    autotiling = []
    path = os.path.join(temp_dir(), "autotiling")
    try:
        for ws in load_string(path).split(","):
            autotiling.append(int(ws))
    except:
        pass
    return autotiling


def num_active_outputs(outputs):
    a = 0
    for output in outputs:
        if output.active:
            a += 1
    return a


def list_outputs(sway=False, silent=False):
    """
    Get output names and geometry from i3 tree, assign to Gdk.Display monitors.
    :return: {"name": str, "x": int, "y": int, "width": int, "height": int, "monitor": Gkd.Monitor}
    """
    outputs_dict = {}
    if sway:
        try:
            from i3ipc import Connection
        except ModuleNotFoundError:
            print("'python-i3ipc' package required on sway, terminating")
            sys.exit(1)

        i3 = Connection()

        if not silent:
            print("Running on sway")
        outputs = i3.get_outputs()
        for item in outputs:
            outputs_dict[item.name] = {"x": item.rect.x,
                                       "y": item.rect.y,
                                       "width": item.rect.width,
                                       "height": item.rect.height,
                                       "description": f"{item.make} {item.model} {item.serial}",
                                       "monitor": None}

    elif os.getenv('HYPRLAND_INSTANCE_SIGNATURE') is not None:
        if not silent:
            print("Running on Hyprland")
        cmd = "hyprctl -j monitors"
        result = subprocess.check_output(cmd, shell=True).decode("utf-8").strip()
        outputs = json.loads(result)
        for item in outputs:
            outputs_dict[item["name"]] = {"x": item["x"],
                                          "y": item["y"],
                                          "width": int(item["width"] / item["scale"]),
                                          "height": int(item["height"] / item["scale"]),
                                          "description": item["description"],
                                          "monitor": None}
            # swap for rotated displays
            if item["transform"] in [1, 3, 5, 7]:
                outputs_dict[item["name"]]["width"] = item["height"]
                outputs_dict[item["name"]]["height"] = item["width"]

    elif os.getenv('NIRI_SOCKET'):
        if not silent:
            print("Running on Niri")
        cmd = "niri msg -j outputs"
        result = subprocess.check_output(cmd, shell=True).decode("utf-8").strip()
        outputs = json.loads(result)
        for item in outputs:
            transform = ""
            x, y, w, h = 0, 0, 0, 0
            for mode in outputs[item]["modes"]:
                if mode["is_preferred"]:
                    x = outputs[item]["logical"]["x"]
                    y = outputs[item]["logical"]["y"]
                    w = mode["width"]
                    h = mode["height"]
                    transform = outputs[item]["logical"]["transform"]
                    break

            outputs_dict[outputs[item]["name"]] = {"x": x,
                                                   "y": y,
                                                   "width": w,
                                                   "height": h,
                                                   "description": f'{outputs[item]["make"]} {outputs[item]["model"]} {outputs[item]["serial"]}',
                                                   "monitor": None}

            # swap for rotated displays
            if transform in ["90", "270", "flipped-90", "flipped-270"]:
                outputs_dict[outputs[item]["name"]]["width"] = h
                outputs_dict[outputs[item]["name"]]["height"] = w

    elif os.getenv('WAYLAND_DISPLAY') is not None:
        if not silent:
            print("Running on Wayland, but neither sway nor Hyprland")
        if nwg_panel.common.commands["wlr-randr"]:
            lines = subprocess.check_output("wlr-randr", shell=True).decode("utf-8").strip().splitlines()
            if lines:
                name, description, w, h, x, y, transform, scale = None, None, None, None, None, None, None, 1.0
                for line in lines:
                    if not line.startswith(" "):
                        name = line.split()[0]
                        description = line.split()[1]
                    elif "current" in line:
                        w_h = line.split()[0].split('x')
                        w = int(w_h[0])
                        h = int(w_h[1])
                    elif "Transform" in line:
                        transform = line.split()[1].strip()
                    elif "Position" in line:
                        x_y = line.split()[1].split(',')
                        x = int(x_y[0])
                        y = int(x_y[1])
                    elif "Scale" in line:
                        try:
                            scale = float(line.split()[1])
                        except ValueError:
                            scale = 1.0

                    if name is not None and w is not None and h is not None and x is not None and y is not None \
                            and transform is not None:
                        if transform == "normal":  # which other values it returns for not rotated displays?
                            outputs_dict[name] = {'name': name,
                                                  'x': x,
                                                  'y': y,
                                                  'width': int(w / scale),
                                                  'height': int(h / scale),
                                                  'transform': transform,
                                                  'description': description,
                                                  'monitor': None}
                        else:
                            outputs_dict[name] = {'name': name,
                                                  'x': x,
                                                  'y': y,
                                                  'width': int(h / scale),
                                                  'height': int(w / scale),
                                                  'transform': transform,
                                                  'description': description,
                                                  'scale': scale,
                                                  'monitor': None}
                        # Each monitor only has a single transform, this is to avoid parsing the same monitor multiple times
                        # Disabled monitors don't have transforms.
                        # Gdk doesn't report disabled monitors, not filtering them would cause crashes
                        transform = None
        else:
            print("'wlr-randr' command not found, terminating")
            sys.exit(1)

    display = Gdk.Display.get_default()
    for i in range(display.get_n_monitors()):
        monitor = display.get_monitor(i)
        geometry = monitor.get_geometry()
        for key in outputs_dict:
            if outputs_dict[key]["x"] == geometry.x and outputs_dict[key]["y"] == geometry.y:
                outputs_dict[key]["monitor"] = monitor
                break

    # map monitor descriptions to output names
    mon_desc2output_name = {}
    for key in outputs_dict:
        if "description" in outputs_dict[key]:
            mon_desc2output_name[outputs_dict[key]["description"]] = key

    return outputs_dict, mon_desc2output_name


def check_key(dictionary, key, default_value):
    """
    Adds a key w/ default value if missing from the dictionary
    """
    if key not in dictionary:
        dictionary[key] = default_value


def cmd2string(cmd):
    process_env = dict(os.environ)
    process_env.update({"LANG": "C.UTF-8"})
    try:
        return subprocess.check_output(cmd, shell=True, env=process_env).decode("utf-8").strip()
    except subprocess.CalledProcessError:
        return ""


# def is_command(cmd):
#     cmd = cmd.split()[0]  # strip arguments
#     cmd = "command -v {}".format(cmd)
#     try:
#         is_cmd = subprocess.check_output(cmd, shell=True).decode("utf-8").strip()
#         if is_cmd:
#             return True
#
#     except subprocess.CalledProcessError:
#         pass
#
#     return False


def is_command(cmd):
    return which(cmd) is not None


def check_commands():
    for key in nwg_panel.common.commands:
        nwg_panel.common.commands[key] = is_command(key)

    try:
        import requests
        nwg_panel.common.commands["python-requests"] = True
    except ModuleNotFoundError:
        pass


def create_background_task(target, interval, args=(), kwargs=None):
    if kwargs is None:
        kwargs = {}

    def loop_wrapper():
        if interval > 0:
            while True:
                target(*args, **kwargs)
                time.sleep(interval)
        else:
            target(*args, **kwargs)

    thread = threading.Thread(target=loop_wrapper, daemon=True)
    return thread


def get_balance(sink_name="@DEFAULT_SINK@"):
    if not nwg_panel.common.commands["pactl"]:
        return 0.0

    try:
        output = subprocess.check_output(
            ["pactl", "get-sink-volume", sink_name], text=True
        )
    except subprocess.CalledProcessError:
        return 0.0

    match = re.search(r"balance\s+(-?\d+[.,]\d+)", output)
    if not match:
        return 0.0

    balance_str = match.group(1).replace(",", ".")
    try:
        return float(balance_str)
    except ValueError:
        return 0.0


def get_volume():
    vol = 0
    muted = False
    if nwg_panel.common.commands["pactl"]:
        try:
            output = cmd2string("pactl get-sink-volume @DEFAULT_SINK@")
            volumes = re.findall(r"/\s+(?P<volume>\d+)%\s+/", output)
            if volumes:
                volumes = [int(x) for x in volumes]
                vol = max(volumes)
        except Exception as e:
            eprint(e)

        try:
            output = cmd2string("pactl get-sink-mute @DEFAULT_SINK@").strip().lower()
            muted = output.endswith("yes")
        except Exception as e:
            eprint(e)
    elif nwg_panel.common.commands["pamixer"]:
        try:
            output = cmd2string("pamixer --get-volume")
            if output:
                vol = int(cmd2string("pamixer --get-volume"))
        except Exception as e:
            eprint(e)

        try:
            muted = subprocess.check_output("pamixer --get-mute", shell=True).decode(
                "utf-8").strip() == "true"
        except subprocess.CalledProcessError:
            # the command above returns the 'disabled` status w/ CalledProcessError, exit status 1
            pass

    else:
        eprint("Couldn't get volume, no 'pamixer' or 'pactl' found")

    return vol, muted


def list_sink_inputs():
    """
    Thanks to @fm16191 for https://github.com/fm16191/pactl-json-parser
    """
    # This used to return localized output, returning unexpected values #327
    # p = subprocess.run("pactl list sink-inputs".split(), capture_output=True)
    # result = p.stdout.decode()

    result = cmd2string("pactl list sink-inputs")

    # Syntax fix
    result = result.replace(u"\xa0", " ")  # Broken encoding for spaces
    result = result.replace(u"\\\"", "\"")  # Replacing all espaced doublequote to inline singlequote (JSON compliant)
    result = result.replace(u"\"", "\'")

    sinks = dict()

    for oline in result.split("\n"):
        if oline == "":
            continue  # Skip empty lines
        # Indentation indicates the JSON structure
        indent = oline.count('\t')
        line = oline[indent:]
        if indent == 0:  # Get sink name
            sink_name = line.split("#")[-1]
            sinks[sink_name] = {}
        elif indent == 1:  # Get sink object
            if line.startswith("        "):  # Output is over two lines
                sinks[sink_name][name] = sinks[sink_name][name] + " " + line.strip()
            else:
                ii = line.index(":")
                name = line[:ii].strip()
                value = line[ii + 1:].strip()
                sinks[sink_name][name] = value
        elif indent == 2:  # Get sink object properties
            if sinks[sink_name][name] == "":
                sinks[sink_name][name] = {}
            ii = line.index("=")
            sub_name = line[:ii].strip()
            sub_value = line[ii + 1:].strip()
            if sub_value[0] == "'" and sub_value[-1] == "'":
                sub_value = sub_value[1:-1]
            sinks[sink_name][name][sub_name] = sub_value
        else:  # Unexpected indentation
            print("Unexpected line : ", oline)
            exit()

    # eprint(json.dumps(sinks, indent=2))
    return sinks


def list_sinks():
    sinks = []
    if nwg_panel.common.commands["pamixer"]:
        try:
            output = cmd2string("pamixer --list-sinks")
            if output:
                lines = output.splitlines()[1:]
                for line in lines:
                    details = line.split()
                    name = details[1][1:-1]
                    desc = " ".join(details[3:])[1:-1]
                    sink = {"name": name, "desc": desc, "running": True if "Running" in line else False}
                    sinks.append(sink)

        except Exception as e:
            eprint(e)

    elif nwg_panel.common.commands["pactl"]:
        try:
            output = cmd2string("pactl list sinks")
            if output:
                lines = output.splitlines()
                sink = {}
                for line in lines:
                    indent = line.count("\t")
                    line = line.lstrip("\t")
                    if indent == 0 and sink:
                        sinks.append(sink)
                        sink = {}
                    elif indent == 1:
                        if line.lower().startswith("name"):
                            sink.update({"name": line.split(": ")[1]})
                        elif line.lower().startswith("description"):
                            sink.update({"desc": line.split(": ")[1]})
                        elif line.lower().startswith("state"):
                            sink.update({"running": True if "RUNNING" in line else False})
                if sink:
                    sinks.append(sink)
        except Exception as e:
            eprint(e)
    else:
        eprint("Couldn't list sinks, no 'pamixer' or 'pactl' found")

    return sinks


def toggle_mute(*args):
    if nwg_panel.common.commands["pamixer"]:
        vol, muted = get_volume()
        if muted:
            subprocess.call("pamixer -u".split())
        else:
            subprocess.call("pamixer -m".split())
    elif nwg_panel.common.commands["pactl"]:
        subprocess.call("pactl set-sink-mute @DEFAULT_SINK@ toggle".split())
    else:
        eprint("Couldn't toggle mute, no 'pamixer' or 'pactl' found")


def set_volume(percent, balance=0):
    if balance < 0:
        left = percent
        right = percent - int(percent * abs(balance))
    elif balance > 0:
        right = percent
        left = percent - int(percent * abs(balance))
    else:
        left = right = percent

    if left < 0:
        left = 0
    if left > 100:
        left = 100

    if right < 0:
        right = 0
    if right > 100:
        right = 100

    if nwg_panel.common.commands["pactl"]:
        subprocess.call(f"pactl set-sink-volume @DEFAULT_SINK@ {left}% {right}%".split())
    elif nwg_panel.common.commands["pamixer"]:
        subprocess.call("pamixer --set-volume {}".format(percent).split())
    else:
        eprint("Couldn't set volume, no 'pamixer' or 'pactl' found")


def get_brightness(device="", controller=""):
    brightness = 0
    if nwg_panel.common.commands["light"] and controller == "light":
        cmd = "light -G -s {}".format(device) if device else "light -G"
        output = cmd2string(cmd)
        brightness = int(round(float(output), 0))
    elif nwg_panel.common.commands["brightnessctl"] and controller == "brightnessctl":
        cmd = "brightnessctl m -d {}".format(device) if device else "brightnessctl m"
        output = cmd2string(cmd)
        max_bri = int(output)

        cmd = "brightnessctl g -d {}".format(device) if device else "brightnessctl g"
        output = cmd2string(cmd)
        b = int(output) * 100 / max_bri
        brightness = int(round(float(b), 0))
    elif nwg_panel.common.commands["ddcutil"] and controller == "ddcutil":
        cmd = "ddcutil getvcp 10 --bus={}".format(device) if device else "ddcutil getvcp 10"
        output = cmd2string(cmd)
        b = int(output.split("current value =")[1].split(",")[0])
        brightness = int(round(float(b), 0))
    else:
        raise ValueError("Couldn't get brightness, is 'light' or 'brightnessctl' or 'ddcutil' installed?")

    return brightness


def set_brightness(percent, device="", controller=""):
    if percent == 0:
        percent = 1
    if nwg_panel.common.commands["light"] and controller == "light":
        if device:
            subprocess.Popen("light -s {} -S {}".format(device, percent).split())
        else:
            subprocess.Popen("light -S {}".format(percent).split())
    elif nwg_panel.common.commands["brightnessctl"] and controller == "brightnessctl":
        if device:
            subprocess.Popen("brightnessctl -d {} s {}%".format(device, percent).split(),
                             stdout=subprocess.DEVNULL,
                             stderr=subprocess.STDOUT)
        else:
            subprocess.Popen("brightnessctl s {}%".format(percent).split(),
                             stdout=subprocess.DEVNULL,
                             stderr=subprocess.STDOUT)
    elif nwg_panel.common.commands["ddcutil"] and controller == "ddcutil":
        if device:
            subprocess.Popen("ddcutil setvcp 10 {} --bus={}".format(percent, device).split())
        else:
            subprocess.Popen("ddcutil setvcp 10 {}".format(percent).split())
    else:
        eprint("Either 'light' or 'brightnessctl' or 'ddcutil' package required")


def get_battery():
    percent, time, charging = 0, "", False
    success = False

    battery_count = len(glob.glob("/sys/class/power_supply/BAT*"))
    if battery_count <= 1:
        try:
            b = psutil.sensors_battery()
            if b:
                percent = int(round(b.percent, 0))
                charging = b.power_plugged
                seconds = b.secsleft
                if seconds != psutil.POWER_TIME_UNLIMITED and seconds != psutil.POWER_TIME_UNKNOWN:
                    time = seconds2string(seconds)
                else:
                    time = ""
                success = True
        except:
            pass

    if not success and nwg_panel.common.commands["upower"]:
        lines = subprocess.check_output(
            "LANG=en_US upower -i $(upower -e | grep devices/DisplayDevice) | grep --color=never -E 'state|to[[:space:]]full|to[[:space:]]empty|percentage'",
            shell=True).decode("utf-8").strip().splitlines()
        for line in lines:
            if "state:" in line:
                charging = line.split(":")[1].strip() == "charging"
            elif "time to" in line:
                time = line.split(":")[1].strip()
            elif "percentage:" in line:
                try:
                    percent = round(float(line.split(":")[1].strip()[:-1]))
                except:
                    pass

    return percent, time, charging


def seconds2string(seconds):
    minutes, sec = divmod(seconds, 60)
    hrs, minutes = divmod(minutes, 60)

    hrs = str(hrs)
    if len(hrs) < 2:
        hrs = "0{}".format(hrs)

    minutes = str(minutes)
    if len(minutes) < 2:
        minutes = "0{}".format(minutes)

    return "{}:{}".format(hrs, minutes)


def update_image(image, icon_name, icon_size, icons_path="", fallback=True):
    scale = image.get_scale_factor()
    icon_size *= scale
    pixbuf = create_pixbuf(icon_name, icon_size, icons_path, fallback)
    surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, scale, image.get_window())
    image.set_from_surface(surface)


def update_image_fallback_desktop(image, icon_name, icon_size, icons_path, fallback=True):
    try:
        # This should work if your icon theme provides the icon, or if it's placed in /usr/share/pixmaps
        update_image(image, icon_name, icon_size, fallback=False)
    except:
        # If the above fails, let's search .desktop files to find the icon name
        icon_from_desktop = get_icon_name(icon_name)
        if icon_from_desktop:
            # trim extension, if given and the definition is not a path
            # if "/" not in icon_from_desktop:
            #     icon_from_desktop = os.path.splitext(icon_from_desktop)[0]

            update_image(image, icon_from_desktop, icon_size, icons_path, fallback=fallback)


def update_gtk_entry(entry, icon_pos, icon_name, icon_size, icons_path=""):
    scale = entry.get_scale_factor()
    icon_size *= scale
    pixbuf = create_pixbuf(icon_name, icon_size, icons_path)
    entry.set_icon_from_pixbuf(icon_pos, pixbuf)


def create_pixbuf(icon_name, icon_size, icons_path="", fallback=True):
    try:
        # In case a full path was given
        if icon_name.startswith("/"):
            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(
                icon_name, icon_size, icon_size)
        else:
            icon_theme = Gtk.IconTheme.get_default()
            if icons_path:
                search_path = icon_theme.get_search_path()
                search_path.append(icons_path)
                icon_theme.set_search_path(search_path)

            try:
                if icons_path:
                    path = "{}/{}.svg".format(icons_path, icon_name)
                    pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(
                        path, icon_size, icon_size)
                else:
                    raise ValueError("icons_path not supplied.")
            except:
                try:
                    pixbuf = icon_theme.load_icon(icon_name, icon_size, Gtk.IconLookupFlags.FORCE_SIZE)
                except:
                    pixbuf = icon_theme.load_icon(icon_name.lower(), icon_size, Gtk.IconLookupFlags.FORCE_SIZE)
    except Exception as e:
        if fallback:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(
                os.path.join(get_config_dir(), "icons_light/icon-missing.svg"), icon_size, icon_size)
        else:
            raise e
    return pixbuf


def list_configs(config_dir):
    configs = {}
    # allow to store json files other than panel config files in the config directory
    # (prevents from crash w/ nwg-drawer>=0.1.7 and future nwg-menu versions)
    exclusions = [os.path.join(config_dir, "preferred-apps.json"),
                  os.path.join(config_dir, "calendar.json"),
                  os.path.join(config_dir, "common-settings.json")]
    entries = os.listdir(config_dir)
    entries.sort()
    for entry in entries:
        path = os.path.join(config_dir, entry)
        if os.path.isfile(path) and path not in exclusions and not path.endswith(".css"):
            try:
                with open(path, 'r') as f:
                    config = json.load(f)
                configs[path] = config
            except:
                pass

    return configs


def get_cache_dir():
    if os.getenv("XDG_CACHE_HOME"):
        return os.getenv("XDG_CACHE_HOME")
    elif os.getenv("HOME") and os.path.isdir(os.path.join(os.getenv("HOME"), ".cache")):
        return os.path.join(os.getenv("HOME"), ".cache")
    else:
        return None


def file_age(path):
    return time.time() - os.stat(path)[stat.ST_MTIME]


def hms():
    return datetime.fromtimestamp(time.time()).strftime("%H:%M:%S")


def get_shell_data_dir():
    data_dir = ""
    home = os.getenv("HOME")
    xdg_data_home = os.getenv("XDG_DATA_HOME")

    if xdg_data_home:
        data_dir = os.path.join(xdg_data_home, "nwg-shell/")
    else:
        if home:
            data_dir = os.path.join(home, ".local/share/nwg-shell/")

    return data_dir


def load_shell_data():
    shell_data_file = os.path.join(get_shell_data_dir(), "data")
    shell_data = load_json(shell_data_file) if os.path.isfile(shell_data_file) else {}

    defaults = {
        "interface-locale": ""
    }

    for key in defaults:
        if key not in shell_data:
            shell_data[key] = defaults[key]

    return shell_data


def niri_ipc(cmd, is_json=False):
    niri_sock = os.getenv("NIRI_SOCKET")
    if not niri_sock:
        print("NIRI_SOCKET environment variable not set.")
        return None

    try:
        # Automatically closes the socket at the end of the with block
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as client:
            client.connect(niri_sock)

            # Send command as stringified JSON or raw JSON
            if not is_json:
                client.send(f'"{cmd}"\n'.encode("utf-8"))
            else:
                client.send(f'{cmd}\n'.encode("utf-8"))

            buffer = ""
            while True:
                chunk = client.recv(1024).decode('utf-8', errors='replace')
                if not chunk:
                    break
                buffer += chunk
                if buffer.endswith('\n'):  # Detect end of message
                    break

    except (socket.error, OSError) as e:
        print("Socket error:", e)
        return None

    try:
        reply = json.loads(buffer)
        key = next(iter(reply))
        return reply[key]
    except json.JSONDecodeError as e:
        print("Failed to decode JSON:", e)
        print("Buffer:", buffer)
        return None


def niri_outputs():
    reply = niri_ipc("Outputs")
    return reply[next(iter(reply))]


def niri_workspaces():
    reply = niri_ipc("Workspaces")
    return reply[next(iter(reply))]


def niri_windows():
    reply = niri_ipc("Windows")
    return reply[next(iter(reply))]


def niri_focused_window():
    reply = niri_ipc("FocusedWindow")
    return reply[next(iter(reply))]


def niri_keyboard_layouts():
    reply = niri_ipc("KeyboardLayouts")
    return reply[next(iter(reply))]


def niri_get_all():
    return niri_outputs(), niri_workspaces(), niri_windows(), niri_focused_window()


def hyprctl(cmd, buf_size=2048):
    # /tmp/hypr moved to $XDG_RUNTIME_DIR/hypr in #5788
    xdg_runtime_dir = os.getenv("XDG_RUNTIME_DIR")
    hypr_dir = f"{xdg_runtime_dir}/hypr" if xdg_runtime_dir and os.path.isdir(
        f"{xdg_runtime_dir}/hypr") else "/tmp/hypr"

    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        s.connect(f"{hypr_dir}/{os.getenv('HYPRLAND_INSTANCE_SIGNATURE')}/.socket.sock")
        s.send(cmd.encode("utf-8"))

        output = b""
        while True:
            buffer = s.recv(buf_size)
            if buffer:
                output = b"".join([output, buffer])
            else:
                break
        s.close()
        return output.decode('utf-8', errors='replace')
    except Exception as e:
        eprint(f"hyprctl: {e}")
        return ""


def h_list_monitors():
    reply = hyprctl("j/monitors")
    try:
        return json.loads(reply)
    except Exception as e:
        eprint(e)
        return {}


def h_list_workspaces():
    reply = hyprctl("j/workspaces")
    try:
        return json.loads(reply)
    except Exception as e:
        eprint(e)
        return {}


def h_list_workspace_rules():
    reply = hyprctl("j/workspacerules")
    try:
        return json.loads(reply)
    except Exception as e:
        eprint(e)
        return {}


def h_list_clients():
    reply = hyprctl("j/clients")
    try:
        return json.loads(reply)
    except Exception as e:
        eprint(e)
        return {}


def h_get_activewindow():
    reply = hyprctl("j/activewindow")
    try:
        return json.loads(reply)
    except Exception as e:
        eprint(e)
        return {}


def h_get_active_workspace():
    reply = hyprctl("j/activeworkspace")
    try:
        return json.loads(reply)
    except Exception as e:
        eprint(e)
        return {}


def h_modules_get_all():
    return h_list_monitors(), h_list_workspaces(), h_list_clients(), h_get_activewindow(), h_get_active_workspace()


def cmd_through_compositor(cmd):
    cs_file = os.path.join(get_config_dir(), "common-settings.json")
    common_settings = load_json(cs_file)

    if "run-through-uwsm" in common_settings and common_settings["run-through-uwsm"] and is_command("uwsm"):
        cmd = f'uwsm app -- {cmd}'
        return cmd

    if "run-through-compositor" not in common_settings or common_settings["run-through-compositor"]:
        cmd = cmd.replace("\"", "\\\"")
        if os.getenv("SWAYSOCK"):
            if os.getenv("XDG_SESSION_DESKTOP") and "miracle-wm" in os.getenv("XDG_SESSION_DESKTOP"):
                cmd = f'miraclemsg exec "{cmd}"'
            else:
                cmd = f'swaymsg exec "{cmd}"'
        elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
            # check if we are on lua dispatchers (Hyprland >= v0.55.0 with lua config)
            res = hyprctl("dispatch hl.dsp.no_op")  # do nothing
            if res == "ok":
                cmd = f"hyprctl dispatch 'hl.dsp.exec_cmd(\"{cmd}\")'"
            else:
                cmd = f'hyprctl dispatch exec "{cmd}"'
        elif os.getenv("NIRI_SOCKET"):
            cmd = f'niri msg action spawn -- {cmd}'

    return cmd


def load_resource(package, resource_name):
    try:
        import importlib.resources as resources
        with resources.open_binary(package, resource_name) as resource_file:
            return resource_file.read()
    except:
        pass

    try:
        import importlib.util
        spec = importlib.util.find_spec(package)
        if spec is not None and spec.loader is not None and hasattr(spec.loader, 'get_data'):
            return spec.loader.get_data(resource_name)
    except Exception:
        pass

    try:
        import pkgutil
        data = pkgutil.get_data(package, resource_name)
        if data is None:
            raise FileNotFoundError(f"Resource {resource_name} not found in package {package}.")
        return data
    except ImportError as e:
        raise ImportError("Failed to load the resource using any available method.") from e



#!/usr/bin/env python3

from gi.repository import Gtk

import os
import subprocess

from nwg_panel.tools import check_key, update_image, cmd_through_compositor


class CustomButton(Gtk.Button):
    def __init__(self, settings, icons_path=""):
        Gtk.Button.__init__(self)
        self.set_always_show_image(True)
        self.settings = settings

        check_key(settings, "icon", "dialog-error")
        check_key(settings, "icon-size", 16)
        image = Gtk.Image()
        update_image(image, settings["icon"], settings["icon-size"], icons_path)
        self.set_image(image)

        check_key(settings, "label", "")
        if "label" in settings and settings["label"]:
            if "label-position" in settings:
                if settings["label-position"] == "right":
                    self.set_image_position(Gtk.PositionType.LEFT)
                    self.set_label(settings["label"])
                elif settings["label-position"] == "left":
                    self.set_image_position(Gtk.PositionType.RIGHT)
                    self.set_label(settings["label"])
                elif settings["label-position"] == "top":
                    self.set_image_position(Gtk.PositionType.BOTTOM)
                    self.set_label(settings["label"])
                elif settings["label-position"] == "bottom":
                    self.set_image_position(Gtk.PositionType.TOP)
                    self.set_label(settings["label"])
                else:
                    self.set_tooltip_text(settings["label"])
            else:
                self.set_tooltip_text(settings["label"])

        if "tooltip" in settings and settings["tooltip"]:
            self.set_tooltip_text(settings["tooltip"])

        check_key(settings, "command", "")
        if settings["command"]:
            self.connect("clicked", self.on_click, settings["command"])

        check_key(settings, "css-name", "")
        if settings["css-name"]:
            self.set_property("name", settings["css-name"])

        self.show()

    def on_click(self, button, cmd):
        if cmd:
            cmd = cmd_through_compositor(cmd)
            print(f"Executing: {cmd}")
            subprocess.Popen('{}'.format(cmd), shell=True)
        else:
            print("No command assigned")
